# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
import logging
import re
import typing
from typing import Any, Dict, List, Optional

from django.db import models
from django.utils.translation import gettext_lazy as _
from pipeline.contrib.node_timer_event.types import TimeDefined
from pipeline.contrib.node_timer_event.utils import parse_timer_defined
from pipeline.core.constants import PE

logger = logging.getLogger(__name__)


EVENT_KEY_PATTERN = re.compile(r".*node:(?P<node_id>.+):version:(?P<version>.+):index:(?P<index>\d+)")


class NodeTimerEventConfigManager(models.Manager):
    def parse_node_timer_event_configs(self, pipeline_tree: Dict[str, Any]) -> Dict[str, Any]:
        """解析事件时间"""
        configs: List[Dict[str, Any]] = []
        for act_id, act in pipeline_tree[PE.activities].items():
            if act["type"] == PE.SubProcess:
                result = self.parse_node_timer_event_configs(act[PE.pipeline])
                if not result["result"]:
                    return result
                configs.extend(result["data"])
            elif act["type"] == PE.ServiceActivity:
                index: int = 1
                treated_timer_events: List[Dict[str, Any]] = []
                timer_events: List[Dict[str, Any]] = (act.get("events") or {}).get("timer_events") or []
                for timer_event in timer_events:
                    enable: bool = timer_event.get("enable") or False
                    if not enable:
                        continue

                    timer_type: Optional[str] = timer_event.get("timer_type")
                    defined: Optional[str] = timer_event.get("defined")

                    try:
                        timer_defined: TimeDefined = parse_timer_defined(timer_type, defined)
                    except Exception:
                        # 对于不符合格式要求的情况，记录日志并跳过
                        logger.exception(
                            "[parse_node_timer_event_configs] parse timer_defined failed: "
                            "node_id -> %s, timer_type -> %s, defined -> %s",
                            act_id,
                            timer_type,
                            defined,
                        )
                        continue

                    treated_timer_events.append(
                        {
                            "index": index,
                            "action": timer_event.get("action"),
                            "timer_type": timer_type,
                            "repetitions": timer_defined["repetitions"],
                            "defined": defined,
                        }
                    )

                    index += 1

                if treated_timer_events:
                    configs.append({"node_id": act_id, "events": treated_timer_events})

        return {"result": True, "data": configs, "message": ""}

    def batch_create_node_timer_event_config(self, root_pipeline_id: str, pipeline_tree: dict):
        """批量创建节点超时配置"""

        config_parse_result: Dict[str, Any] = self.parse_node_timer_event_configs(pipeline_tree)
        # 这里忽略解析失败的情况，保证即使解析失败也能正常创建任务
        if not config_parse_result["result"]:
            logger.error(
                f"[batch_create_node_timer_event_config] parse node timer event config "
                f'failed: {config_parse_result["result"]}'
            )
            return config_parse_result

        configs: List[Dict[str, Any]] = config_parse_result["data"]
        config_objs: typing.List[NodeTimerEventConfig] = [
            NodeTimerEventConfig(
                root_pipeline_id=root_pipeline_id, node_id=config["node_id"], events=json.dumps(config["events"])
            )
            for config in configs
        ]
        objs = self.bulk_create(config_objs, batch_size=1000)
        return {"result": True, "data": objs, "message": ""}


class NodeTimerEventConfig(models.Model):
    root_pipeline_id = models.CharField(verbose_name="root pipeline id", max_length=64)
    node_id = models.CharField(verbose_name="task node id", max_length=64, primary_key=True)
    events = models.TextField(verbose_name="timer events", default="[]")

    objects = NodeTimerEventConfigManager()

    class Meta:
        verbose_name = _("节点时间事件配置 NodeTimerEventConfig")
        verbose_name_plural = _("节点时间事件配置 NodeTimerEventConfig")
        index_together = [("root_pipeline_id", "node_id")]

    def get_events(self) -> List[Dict[str, Any]]:
        return json.loads(self.events)


class ExpiredNodesRecord(models.Model):
    id = models.BigAutoField(verbose_name="ID", primary_key=True)
    nodes = models.TextField(verbose_name="到期节点信息")

    class Meta:
        verbose_name = _("到期节点数据记录 ExpiredNodesRecord")
        verbose_name_plural = _("到期节点数据记录 ExpiredNodesRecord")



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import models
from django.utils.translation import gettext_lazy as _
from pipeline.contrib.fields import SerializerField


class ScheduleManger(models.Manager):
    def apply_schedule_lock(self, task_id: int) -> bool:
        """
        获取 Schedule 对象的调度锁，返回是否成功获取锁

        :return: True or False
        """
        return self.filter(id=task_id, scheduling=False).update(scheduling=True) == 1

    def release_schedule_lock(self, task_id: int) -> None:
        """
        释放指定 Schedule 的调度锁
        :return:
        """
        self.filter(id=task_id, scheduling=True).update(scheduling=False)


class ScheduleLock(object):
    def __init__(self, task_id: int):
        self.task_id = task_id
        self.locked = False

    def __enter__(self):
        self.locked = PluginExecuteTask.objects.apply_schedule_lock(self.task_id)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.locked:
            PluginExecuteTask.objects.release_schedule_lock(self.task_id)


def get_schedule_lock(task_id: int) -> ScheduleLock:
    """
    获取 schedule lock 的 context 对象
    :param task_id:
    :return:
    """
    return ScheduleLock(task_id)


class PluginExecuteTask(models.Model):
    """
    单节点执行任务
    """

    state = models.CharField(_("状态名"), null=False, max_length=64)
    invoke_count = models.IntegerField("invoke count", default=1)
    component_code = models.CharField(_("组件编码"), max_length=255, db_index=True)
    version = models.CharField(_("插件版本"), max_length=255, default="legacy")
    inputs = SerializerField(verbose_name=_("node inputs"), default={})
    outputs = SerializerField(verbose_name=_("node outputs"), default={})
    callback_data = SerializerField(verbose_name=_("callback data"), default={})
    contexts = SerializerField(verbose_name=_("pipeline context values"), default={})
    runtime_attrs = SerializerField(verbose_name=_("runtime attr"), default={})
    scheduling = models.BooleanField("是否正在调度", default=False, db_index=True)
    created_at = models.DateTimeField("create time", auto_now_add=True)
    finish_at = models.DateTimeField("finish time", null=True)

    objects = ScheduleManger()

    class Meta:
        indexes = [models.Index(fields=["id", "scheduling"])]



# Generated by Django 3.2.18 on 2023-10-25 02:06

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ExpiredNodesRecord",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("nodes", models.TextField(verbose_name="到期节点信息")),
            ],
            options={
                "verbose_name": "到期节点数据记录 ExpiredNodesRecord",
                "verbose_name_plural": "到期节点数据记录 ExpiredNodesRecord",
            },
        ),
        migrations.CreateModel(
            name="NodeTimerEventConfig",
            fields=[
                ("root_pipeline_id", models.CharField(max_length=64, verbose_name="root pipeline id")),
                (
                    "node_id",
                    models.CharField(max_length=64, primary_key=True, serialize=False, verbose_name="task node id"),
                ),
                ("events", models.TextField(default="[]", verbose_name="timer events")),
            ],
            options={
                "verbose_name": "节点时间事件配置 NodeTimerEventConfig",
                "verbose_name_plural": "节点时间事件配置 NodeTimerEventConfig",
                "index_together": {("root_pipeline_id", "node_id")},
            },
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import logging
from typing import List, Optional, Type

from django.dispatch import receiver
from pipeline.contrib.node_timer_event.adapter import NodeTimerEventBaseAdapter
from pipeline.contrib.node_timer_event.settings import node_timer_event_settings
from pipeline.eri.signals import post_set_state

from bamboo_engine import states as bamboo_engine_states

logger = logging.getLogger(__name__)


def _node_timer_event_info_update(redis_inst, to_state: str, node_id: str, version: str):

    adapter: Optional[NodeTimerEventBaseAdapter] = None

    if to_state in [
        bamboo_engine_states.RUNNING,
        bamboo_engine_states.FAILED,
        bamboo_engine_states.FINISHED,
        bamboo_engine_states.SUSPENDED,
    ]:

        adapter_class: Type[NodeTimerEventBaseAdapter] = node_timer_event_settings.adapter_class
        adapter: NodeTimerEventBaseAdapter = adapter_class(node_id=node_id, version=version)

        if not adapter.is_ready():
            logger.info(
                "[node_timer_event_info_update] node_timer_event_config not exist and skipped: "
                "node_id -> %s, version -> %s",
                node_id,
                version,
            )
            return

        logger.info(
            "[node_timer_event_info_update] load node_timer_event_config: node_id -> %s, version -> %s, events -> %s",
            node_id,
            version,
            adapter.events,
        )

    if to_state == bamboo_engine_states.RUNNING:
        # 遍历节点时间事件，丢进待调度节点池
        for event in adapter.events:
            adapter.add_to_pool(redis_inst, event)

    elif to_state in [bamboo_engine_states.FAILED, bamboo_engine_states.FINISHED, bamboo_engine_states.SUSPENDED]:
        keys: List[str] = adapter.fetch_keys_to_be_rem()
        redis_inst.zrem(node_timer_event_settings.executing_pool, *keys)
        redis_inst.delete(*keys)
        logger.info(
            "[node_timer_event_info_update] removed events from redis: "
            "node_id -> %s, version -> %s, events -> %s, keys -> %s",
            node_id,
            version,
            adapter.events,
            keys,
        )


@receiver(post_set_state)
def bamboo_engine_eri_node_state_handler(sender, node_id, to_state, version, root_id, parent_id, loop, **kwargs):
    try:
        _node_timer_event_info_update(node_timer_event_settings.redis_inst, to_state, node_id, version)
    except Exception:
        logger.exception("node_timeout_info_update error")



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json
import re

from django.db import transaction
from pipeline.contrib.exceptions import UpdatePipelineContextException
from pipeline.eri.models import ContextValue
from pipeline.eri.models import ExecutionData as DBExecutionData
from pipeline.eri.models import State
from pipeline.eri.runtime import BambooDjangoRuntime

from bamboo_engine import api, states
from bamboo_engine.eri import ContextValue as ContextValueInfo
from bamboo_engine.eri import ContextValueType

FORMATTED_KEY_PATTERN = re.compile(r"^\${(.*?)}$")


class MockHandler:
    def __init__(self, root_pipeline_id, node_id, context_values):
        self.root_pipeline_id = root_pipeline_id
        self.node_id = node_id
        self.context_values = context_values

    def update_node_outputs(self):
        """
        批量修改任务某个节点的输出和上下文
        :param root_pipeline_id: pipeline的id
        :param node_id: 节点id
        :param context_values: {
            "${code}": 200
        }
        :return:
        """

        pipeline_state = State.objects.filter(node_id=self.root_pipeline_id).first()
        if not pipeline_state:
            raise UpdatePipelineContextException(
                "update context values failed: pipeline state not exist, root_pipeline_id={}".format(
                    self.root_pipeline_id
                )
            )

        if pipeline_state.name != states.RUNNING:
            raise UpdatePipelineContextException(
                "update context values failed: the task of non-running state is not allowed update, "
                "root_pipeline_id={}".format(self.root_pipeline_id)
            )

        node_state = State.objects.filter(node_id=self.node_id).first()
        if not node_state:
            raise UpdatePipelineContextException(
                "update context values failed: node state not exist, root_pipeline_id={}".format(self.root_pipeline_id)
            )

        if node_state.name != states.FAILED:
            raise UpdatePipelineContextException(
                "update context values failed: \
              the task of non-failed state is not allowed to update, node_id={}".format(
                    self.node_id
                )
            )

        if "${_system}" in self.context_values.keys():
            raise UpdatePipelineContextException("${_system} is built-in variable that is not allowed to be updated")

        # 获取流程内满足上下文的key
        context_value_queryset = ContextValue.objects.filter(
            pipeline_id=self.root_pipeline_id, key__in=self.context_values.keys()
        )
        context_value_list = []

        for context_value in context_value_queryset:
            if context_value.key in self.context_values.keys():
                context_value_list.append(
                    ContextValueInfo(
                        key=context_value.key,
                        type=ContextValueType(context_value.type),
                        value=self.context_values.get(context_value.key),
                        code=context_value.code,
                    )
                )

        with transaction.atomic():
            try:
                runtime = BambooDjangoRuntime()
                api.update_context_values(runtime, self.root_pipeline_id, context_value_list)
            except Exception as e:
                raise UpdatePipelineContextException("update context value failed, please check it, error={}".format(e))

            outputs = {}
            try:
                for key, value in self.context_values.items():
                    if FORMATTED_KEY_PATTERN.match(key):
                        key = key[2:-1]
                    outputs[key] = value
                execution_data = DBExecutionData.objects.get(node_id=self.node_id)
                detail = json.loads(execution_data.outputs)
                detail.update(outputs)
                execution_data.outputs = json.dumps(detail)
                execution_data.save()

            except Exception as e:
                raise UpdatePipelineContextException(
                    "update node outputs value failed, please check it,outputs={}, error={}".format(outputs, e)
                )



# Generated by Django 3.2.23 on 2023-12-29 09:11

import pipeline.contrib.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("rollback", "0002_auto_20231020_1234"),
    ]

    operations = [
        migrations.AddField(
            model_name="rollbackplan",
            name="options",
            field=pipeline.contrib.fields.SerializerField(default={}, verbose_name="rollback options"),
        ),
    ]



# Generated by Django 3.2.18 on 2023-10-20 12:34

import pipeline.contrib.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("rollback", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="rollbacknodesnapshot",
            name="context_values",
            field=pipeline.contrib.fields.SerializerField(verbose_name="pipeline context values"),
        ),
        migrations.AlterField(
            model_name="rollbacknodesnapshot",
            name="inputs",
            field=pipeline.contrib.fields.SerializerField(verbose_name="node inputs"),
        ),
        migrations.AlterField(
            model_name="rollbacknodesnapshot",
            name="outputs",
            field=pipeline.contrib.fields.SerializerField(verbose_name="node outputs"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from pipeline.service import task_service
from .base import ActionRequestHandler, HandlerType


class PipelineEngineRequestHandler(ActionRequestHandler):
    handler_type = HandlerType.PIPELINE_ENGINE

    def task_pause(self):
        return task_service.pause_pipeline(pipeline_id=self.instance_id)

    def task_resume(self, **kwargs):
        return task_service.resume_pipeline(pipeline_id=self.instance_id)

    def task_revoke(self, **kwargs):
        return task_service.revoke_pipeline(pipeline_id=self.instance_id)

    def node_retry(self, **kwargs):
        return task_service.retry_activity(act_id=self.instance_id, inputs=kwargs.get("inputs"))

    def node_skip(self, **kwargs):
        return task_service.skip_activity(act_id=self.instance_id)

    def node_callback(self, **kwargs):
        return task_service.callback(act_id=self.instance_id, data=kwargs.get("data"))

    def node_skip_exg(self, **kwargs):
        return task_service.skip_exclusive_gateway(gateway_id=self.instance_id, flow_id=kwargs["flow_id"])

    def node_skip_cpg(self, **kwargs):
        return task_service.skip_conditional_parallel_gateway(
            gateway_id=self.instance_id, flow_ids=kwargs["flow_ids"], converge_gateway_id=kwargs["converge_gateway_id"]
        )

    def node_forced_fail(self, **kwargs):
        return task_service.forced_fail(
            act_id=self.instance_id, ex_data=f"force fail by {kwargs.get('operator', 'engine_admin')}"
        )



# -*- coding: utf-8 -*-
from django.db import models
from django.utils.translation import gettext_lazy as _
from pipeline.contrib.fields import SerializerField
from pipeline.contrib.rollback.constants import TOKEN


class RollbackToken(models.Model):
    """
    回滚配置token信息
    """

    root_pipeline_id = models.CharField(verbose_name="root pipeline id", max_length=64, db_index=True)
    token = models.TextField(_("token map"), null=False)
    is_deleted = models.BooleanField(_("is deleted"), default=False, help_text=_("is deleted"), db_index=True)


class RollbackSnapshot(models.Model):
    """
    节点执行的快照信息
    """

    root_pipeline_id = models.CharField(verbose_name="root pipeline id", max_length=64, db_index=True)
    graph = models.TextField(verbose_name="rollback graph", null=False)
    node_access_record = models.TextField(verbose_name="node access record")
    skip_rollback_nodes = models.TextField(verbose_name="skip rollback nodes")
    other_nodes = models.TextField(verbose_name="other nodes")
    start_node_id = models.CharField(verbose_name="start node id", max_length=64, db_index=True)
    target_node_id = models.CharField(verbose_name="target_node_id", max_length=64, db_index=True)
    is_expired = models.BooleanField(verbose_name="is expired", default=False, db_index=True)


class RollbackNodeSnapshot(models.Model):
    """
    节点快照
    """

    root_pipeline_id = models.CharField(verbose_name="root pipeline id", max_length=64, db_index=True)
    node_id = models.CharField(verbose_name="node_id", max_length=64, db_index=True)
    code = models.CharField(verbose_name="node_code", max_length=64)
    version = models.CharField(verbose_name=_("version"), null=False, max_length=33)
    inputs = SerializerField(verbose_name=_("node inputs"))
    outputs = SerializerField(verbose_name=_("node outputs"))
    context_values = SerializerField(verbose_name=_("pipeline context values"))
    rolled_back = models.BooleanField(_("whether the node rolls back"), default=False)


class RollbackPlan(models.Model):
    root_pipeline_id = models.CharField(verbose_name="root pipeline id", max_length=64, db_index=True)
    start_node_id = models.CharField(verbose_name="start node id", max_length=64, db_index=True)
    target_node_id = models.CharField(verbose_name="target_node_id", max_length=64, db_index=True)
    mode = models.CharField(verbose_name="rollback mode", max_length=32, default=TOKEN)
    options = SerializerField(verbose_name="rollback options", default={})
    is_expired = models.BooleanField(verbose_name="is expired", default=False)



# -*- coding: utf-8 -*-
import copy

from pipeline.contrib.rollback import constants
from pipeline.core.constants import PE

from bamboo_engine.utils.graph import RollbackGraph


class CycleHandler:
    """
    环处理器，作用是去除拓扑中的环
    """

    def __init__(self, node_map):
        self.node_map = copy.deepcopy(node_map)

    def get_nodes_and_edges(self):
        """
        从node_map 中解析出环和边
        """
        nodes = []
        edges = []
        for node, value in self.node_map.items():
            nodes.append(node)
            targets = value["targets"]
            for target in targets.values():
                # 过滤掉那些没有执行的分支
                if target not in self.node_map:
                    continue
                edges.append([node, target])
        return nodes, edges

    def has_cycle(self, nodes, edges) -> list:
        """
        判断是否有环，存在环是，将返回一个有效的list
        """
        graph = RollbackGraph(nodes, edges)
        return graph.get_cycle()

    def delete_edge(self, source, target):
        """
        删除环边
        """
        targets = self.node_map[source]["targets"]

        keys_to_remove = []
        for key, val in targets.items():
            if val == target:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            del targets[key]

    def remove_cycle(self):
        while True:
            nodes, edges = self.get_nodes_and_edges()
            cycles = self.has_cycle(nodes, edges)
            if not cycles:
                break
            source = cycles[-2]
            target = cycles[-1]
            self.delete_edge(source, target)
        return self.node_map


class RollbackGraphHandler:
    def __init__(self, node_map, start_id, target_id):
        self.graph = RollbackGraph()
        # 回滚开始的节点
        self.start_id = start_id
        # 回滚结束的节点
        self.target_id = target_id
        self.graph.add_node(start_id)
        self.graph.add_node(target_id)
        # 去除自环边
        self.node_map = CycleHandler(node_map).remove_cycle()
        # 其他不参与回滚，但是需要被清理的节点，主要是网关节点和子流程节点
        self.others_nodes = []

    def build(self, node_id, source_id=None):
        """
        使用递归构建用于回滚的图谱，最终会生成一条连线 source_id -> node_id
        @param node_id 本次遍历到的节点id
        @param source_id 上一个遍历到的节点id
        """
        node_detail = self.node_map.get(node_id)
        if node_detail is None:
            return
        node_type = node_detail["type"]

        if node_type not in [PE.ServiceActivity]:
            self.others_nodes.append(node_id)

        if node_type == PE.ServiceActivity:
            next_node_id = node_detail.get("id")
            self.graph.add_node(next_node_id)
            if source_id and source_id != next_node_id:
                self.graph.add_edge(source_id, next_node_id)

            # 如果遍历到目标节点，则返回
            if node_id == self.start_id:
                return
            source_id = next_node_id
            targets = node_detail.get("targets", {}).values()
        elif node_type == PE.SubProcess:
            # 处理子流程
            source_id = self.build(node_detail["start_event_id"], source_id)
            targets = node_detail.get("targets", {}).values()
        elif node_type == PE.ExclusiveGateway:
            targets = [target for target in node_detail.get("targets", {}).values() if target in self.node_map.keys()]
        else:
            targets = node_detail.get("targets", {}).values()

        # 为了避免循环的过程中source_id值被覆盖，需要额外临时存储source_id
        temporary_source_id = source_id
        for target in targets:
            source_id = self.build(target, temporary_source_id)

        return source_id

    def build_rollback_graph(self):
        """
        这里将会从结束的节点往开始的节点进行遍历，之后再反转图
        """
        self.graph.add_node(constants.END_FLAG)
        # 未整个流程加上结束节点
        self.graph.add_edge(constants.END_FLAG, self.target_id)
        self.graph.add_node(constants.START_FLAG)
        self.graph.add_edge(self.start_id, constants.START_FLAG)
        self.build(self.target_id, self.target_id)

        return self.graph.reverse(), self.others_nodes



# -*- coding: utf-8 -*-
import codecs
import json
import pickle

from django.db.models import TextField


class SerializerField(TextField):
    """
    特定的序列化类，用于兼容json和pickle两种序列化数据
    """

    def to_python(self, value):
        try:
            return json.loads(value)
        except Exception:
            return pickle.loads(codecs.decode(value.encode(), "base64"))

    def from_db_value(self, value, expression, connection, context=None):
        try:
            return json.loads(value)
        except Exception:
            return pickle.loads(codecs.decode(value.encode(), "base64"))

    def get_prep_value(self, value):
        try:
            return json.dumps(value)
        except TypeError:
            return codecs.encode(pickle.dumps(value), "base64").decode()
        pass



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from pipeline.contrib.rollback.constants import TOKEN
from pipeline.contrib.rollback.handler import RollbackDispatcher
from pipeline.contrib.utils import ensure_return_pipeline_contrib_api_result


@ensure_return_pipeline_contrib_api_result
def rollback(
    root_pipeline_id: str,
    start_node_id: str,
    target_node_id: str,
    skip_rollback_nodes: list = None,
    mode: str = TOKEN,
    **options
):
    """
    :param root_pipeline_id: pipeline id
    :param start_node_id: 开始的 id
    :param target_node_id: 开始的 id
    :param skip_rollback_nodes: 需要跳过回滚的节点信息， 仅在TOKEN模式下有效
    :param skip_check_token: 是否跳过检查，仅在ANY模式下有效
    :param mode 回滚模式
    :return: True or False
    """
    RollbackDispatcher(root_pipeline_id, mode).rollback(
        start_node_id, target_node_id, skip_rollback_nodes=skip_rollback_nodes, **options
    )


@ensure_return_pipeline_contrib_api_result
def reserve_rollback(root_pipeline_id: str, start_node_id: str, target_node_id: str, mode: str = TOKEN, **options):
    """
    预约一次回滚任务
    :param root_pipeline_id: pipeline id
    :param start_node_id: 回滚的起始id
    :param target_node_id: 回滚的目标节点id
    :param mode: 模式
    :param skip_check_token: 是否跳过检查，仅在ANY模式下有效
    :return: True or False
    """
    RollbackDispatcher(root_pipeline_id, mode).reserve_rollback(start_node_id, target_node_id, **options)


@ensure_return_pipeline_contrib_api_result
def cancel_reserved_rollback(root_pipeline_id: str, start_node_id: str, target_node_id: str, mode: str = TOKEN):
    """
    取消预约的回滚任务
    :param root_pipeline_id: pipeline id
    :param start_node_id: 回滚的起始id
    :param target_node_id: 回滚的目标节点id
    :param mode: 模式
    :return: True or False
    """
    RollbackDispatcher(root_pipeline_id, mode).cancel_reserved_rollback(start_node_id, target_node_id)


@ensure_return_pipeline_contrib_api_result
def retry_rollback_failed_node(root_pipeline_id: str, node_id: str, retry_data: dict = None, mode: str = TOKEN):
    """
    重试回滚, 仅支持token模式下的回滚重试
    :param root_pipeline_id: pipeline id
    :param node_id: 要重试的节点id
    :param retry_data: 重试的数据
    :param mode: 回滚模式
    """
    RollbackDispatcher(root_pipeline_id, mode).retry_rollback_failed_node(node_id, retry_data)


@ensure_return_pipeline_contrib_api_result
def get_allowed_rollback_node_id_list(root_pipeline_id: str, start_node_id: str, mode: str = TOKEN, **options):
    """
    获取允许回滚的节点范围
    :param root_pipeline_id: pipeline id
    :param start_node_id: 回滚的开始位置
    :param mode: 回滚的模式
    :param skip_check_token: 是否跳过检查，仅在ANY模式下有效
    """
    return RollbackDispatcher(root_pipeline_id, mode).get_allowed_rollback_node_id_list(start_node_id, **options)



# -*- coding: utf-8 -*-
import json
import logging

from celery import current_app
from django.conf import settings
from django.db import transaction
from pipeline.conf.default_settings import ROLLBACK_QUEUE
from pipeline.contrib.rollback import constants
from pipeline.contrib.rollback.models import (
    RollbackNodeSnapshot,
    RollbackPlan,
    RollbackSnapshot,
)
from pipeline.eri.models import CallbackData
from pipeline.eri.models import ExecutionData as DBExecutionData
from pipeline.eri.models import (
    ExecutionHistory,
    LogEntry,
    Node,
    Process,
    Schedule,
    State,
)
from pipeline.eri.runtime import BambooDjangoRuntime

from bamboo_engine import states
from bamboo_engine.eri import ExecutionData
from bamboo_engine.utils.graph import RollbackGraph

logger = logging.getLogger("celery")


class RollbackCleaner:
    def __init__(self, snapshot):
        self.snapshot = snapshot

    def _clear_node_reserve_flag(self, node_id):
        node = Node.objects.get(node_id=node_id)
        node_detail = json.loads(node.detail)
        node_detail["reserve_rollback"] = False
        node.detail = json.dumps(node_detail)
        node.save()

    def clear_data(self):
        # 节点快照需要全部删除，可能会有下一次的回滚
        RollbackNodeSnapshot.objects.filter(root_pipeline_id=self.snapshot.root_pipeline_id).delete()
        # 回滚快照需要置为已过期
        RollbackSnapshot.objects.filter(root_pipeline_id=self.snapshot.root_pipeline_id).update(is_expired=True)
        # 预约计划需要修改为已过期
        RollbackPlan.objects.filter(
            root_pipeline_id=self.snapshot.root_pipeline_id, start_node_id=self.snapshot.start_node_id
        ).update(is_expired=True)
        # 节点的预约信息需要清理掉
        self._clear_node_reserve_flag(self.snapshot.start_node_id)
        # 需要删除该节点的进程信息/非主进程的，防止网关再分支处回滚时，仍然有正在运行的process得不到清理
        Process.objects.filter(
            root_pipeline_id=self.snapshot.root_pipeline_id, current_node_id=self.snapshot.start_node_id
        ).exclude(parent_id=-1).delete()

        graph = json.loads(self.snapshot.graph)
        need_clean_node_id_list = graph["nodes"] + json.loads(self.snapshot.other_nodes)
        # 清理状态信息
        State.objects.filter(root_id=self.snapshot.root_pipeline_id, node_id__in=need_clean_node_id_list).delete()
        # 清理Schedule 信息
        Schedule.objects.filter(node_id__in=need_clean_node_id_list).delete()
        # 清理日志信息
        LogEntry.objects.filter(node_id__in=need_clean_node_id_list).delete()
        ExecutionHistory.objects.filter(node_id__in=need_clean_node_id_list).delete()
        DBExecutionData.objects.filter(node_id__in=need_clean_node_id_list).delete()
        CallbackData.objects.filter(node_id__in=need_clean_node_id_list).delete()


class TokenRollbackTaskHandler:
    def __init__(self, snapshot_id, node_id, retry, retry_data):
        self.snapshot_id = snapshot_id
        self.node_id = node_id
        self.retry_data = retry_data
        self.retry = retry
        self.runtime = BambooDjangoRuntime()

    def set_state(self, node_id, state):
        logger.info("[TokenRollbackTaskHandler][set_state] set node_id state to {}".format(state))
        # 开始和结束节点直接跳过回滚
        if node_id in [constants.END_FLAG, constants.START_FLAG]:
            return
        self.runtime.set_state(
            node_id=node_id,
            to_state=state,
        )

    def execute_rollback(self):
        """
        执行回滚的操作
        """
        if self.node_id in [constants.END_FLAG, constants.START_FLAG]:
            return True

        # 获取节点快照，可能会有多多份快照，需要多次回滚
        node_snapshots = RollbackNodeSnapshot.objects.filter(node_id=self.node_id, rolled_back=False).order_by("-id")
        for node_snapshot in node_snapshots:
            service = self.runtime.get_service(code=node_snapshot.code, version=node_snapshot.version)
            data = ExecutionData(inputs=node_snapshot.inputs, outputs=node_snapshot.outputs)
            parent_data = ExecutionData(inputs=node_snapshot.context_values, outputs={})
            result = service.service.rollback(data, parent_data, self.retry_data)
            node_snapshot.rolled_back = True
            node_snapshot.save()
            if not result:
                return False

        return True

    def start_pipeline(self, root_pipeline_id, target_node_id):
        """
        启动pipeline
        """
        # 将当前住进程的正在运行的节点指向目标ID
        main_process = Process.objects.get(root_pipeline_id=root_pipeline_id, parent_id=-1)
        main_process.current_node_id = target_node_id
        main_process.save()

        # 重置该节点的状态信息
        self.runtime.set_state(
            node_id=target_node_id,
            to_state=states.READY,
            is_retry=True,
            refresh_version=True,
            clear_started_time=True,
            clear_archived_time=True,
        )

        process_info = self.runtime.get_process_info(main_process.id)
        # 设置pipeline的状态
        self.runtime.set_state(
            node_id=root_pipeline_id,
            to_state=states.READY,
        )

        # 如果开启了流程自动回滚，则会开启到目标节点之后自动开始
        if getattr(settings, "PIPELINE_ENABLE_AUTO_EXECUTE_WHEN_ROLL_BACKED", True):
            self.runtime.set_state(
                node_id=root_pipeline_id,
                to_state=states.RUNNING,
            )
        else:
            # 流程设置为暂停状态，需要用户点击才可以继续开始
            self.runtime.set_state(
                node_id=root_pipeline_id,
                to_state=states.SUSPENDED,
            )

        self.runtime.execute(
            process_id=process_info.process_id,
            node_id=target_node_id,
            root_pipeline_id=process_info.root_pipeline_id,
            parent_pipeline_id=process_info.top_pipeline_id,
        )

    def rollback(self):
        with transaction.atomic():
            rollback_snapshot = RollbackSnapshot.objects.select_for_update().get(id=self.snapshot_id, is_expired=False)
            node_access_record = json.loads(rollback_snapshot.node_access_record)
            # 只有非重试状态下才需要记录访问
            if not self.retry:
                node_access_record[self.node_id] += 1
                rollback_snapshot.node_access_record = json.dumps(node_access_record)
                rollback_snapshot.save()

        graph = json.loads(rollback_snapshot.graph)
        target_node_id = rollback_snapshot.target_node_id
        rollback_graph = RollbackGraph(nodes=graph["nodes"], flows=graph["flows"])
        skip_rollback_nodes = json.loads(rollback_snapshot.skip_rollback_nodes)
        in_degrees = rollback_graph.in_degrees()

        clearner = RollbackCleaner(rollback_snapshot)

        if node_access_record[self.node_id] >= in_degrees[self.node_id]:
            # 对于不需要跳过的节点才会执行具体的回滚行为
            if self.node_id not in skip_rollback_nodes:
                try:
                    # 设置节点状态为回滚中
                    self.set_state(self.node_id, states.ROLLING_BACK)
                    # 执行同步回滚的操作
                    result = self.execute_rollback()
                except Exception as e:
                    logger.error(
                        "[TokenRollbackTaskHandler][rollback] execute rollback error,"
                        "snapshot_id={}, node_id={}, err={}".format(self.snapshot_id, self.node_id, e)
                    )
                    # 节点和流程重置为回滚失败的状态
                    self.set_state(rollback_snapshot.root_pipeline_id, states.ROLL_BACK_FAILED)
                    # 回滚失败的节点将不再向下执行
                    self.set_state(self.node_id, states.ROLL_BACK_FAILED)
                    return

                # 节点回滚成功
                if result:
                    self.set_state(self.node_id, states.ROLL_BACK_SUCCESS)
                else:
                    logger.info(
                        "[TokenRollbackTaskHandler][rollback], execute rollback failed, "
                        "result=False, snapshot_id={}, node_id={}".format(self.snapshot_id, self.node_id)
                    )
                    self.set_state(self.node_id, states.ROLL_BACK_FAILED)
                    # 回滚失败的节点将不再向下执行
                    self.set_state(rollback_snapshot.root_pipeline_id, states.ROLL_BACK_FAILED)
                    return

            next_node = rollback_graph.next(self.node_id)
            if list(next_node)[0] == constants.END_FLAG:
                self.set_state(rollback_snapshot.root_pipeline_id, states.ROLL_BACK_SUCCESS)
                try:
                    clearner.clear_data()
                    self.start_pipeline(
                        root_pipeline_id=rollback_snapshot.root_pipeline_id, target_node_id=target_node_id
                    )
                except Exception as e:
                    logger.error("[TokenRollbackTaskHandler][rollback] start_pipeline failed, err={}".format(e))
                    return
                return

            for node in next_node:
                token_rollback.apply_async(
                    kwargs={
                        "snapshot_id": self.snapshot_id,
                        "node_id": node,
                    },
                    queue=ROLLBACK_QUEUE,
                )


class AnyRollbackHandler:
    def __init__(self, snapshot_id):
        self.snapshot_id = snapshot_id
        self.runtime = BambooDjangoRuntime()

    def start_pipeline(self, root_pipeline_id, target_node_id):
        """
        启动pipeline
        """

        main_process = Process.objects.get(root_pipeline_id=root_pipeline_id, parent_id=-1)
        main_process.current_node_id = target_node_id
        main_process.save()

        # 重置该节点的状态信息
        self.runtime.set_state(
            node_id=target_node_id,
            to_state=states.READY,
            is_retry=True,
            refresh_version=True,
            clear_started_time=True,
            clear_archived_time=True,
        )

        process_info = self.runtime.get_process_info(main_process.id)

        # 如果PIPELINE_ENABLE_AUTO_EXECUTE_WHEN_ROLL_BACKED为False, 那么则会重制流程为暂停状态
        if not getattr(settings, "PIPELINE_ENABLE_AUTO_EXECUTE_WHEN_ROLL_BACKED", True):
            self.runtime.set_state(
                node_id=root_pipeline_id,
                to_state=states.SUSPENDED,
            )

        self.runtime.execute(
            process_id=process_info.process_id,
            node_id=target_node_id,
            root_pipeline_id=process_info.root_pipeline_id,
            parent_pipeline_id=process_info.top_pipeline_id,
        )

    def rollback(self):
        with transaction.atomic():
            rollback_snapshot = RollbackSnapshot.objects.get(id=self.snapshot_id, is_expired=False)
            clearner = RollbackCleaner(rollback_snapshot)
            try:
                clearner.clear_data()
                self.start_pipeline(
                    root_pipeline_id=rollback_snapshot.root_pipeline_id, target_node_id=rollback_snapshot.target_node_id
                )
            except Exception as e:
                logger.error(
                    "rollback failed: start pipeline, pipeline_id={}, target_node_id={}, error={}".format(
                        rollback_snapshot.root_pipeline_id, rollback_snapshot.target_node_id, str(e)
                    )
                )
                raise e


@current_app.task
def token_rollback(snapshot_id, node_id, retry=False, retry_data=None):
    """
    snapshot_id 本次回滚的快照id
    node_id  当前要回滚的节点id
    """
    TokenRollbackTaskHandler(snapshot_id=snapshot_id, node_id=node_id, retry=retry, retry_data=retry_data).rollback()


@current_app.task
def any_rollback(snapshot_id):
    AnyRollbackHandler(snapshot_id=snapshot_id).rollback()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import datetime
import re
from typing import Any, Callable, Dict, Optional

import isodate
from pipeline.contrib.node_timer_event.constants import TimerType
from pipeline.contrib.node_timer_event.types import TimeDefined

TIME_CYCLE_DEFINED_PATTERN = re.compile(r"(^R\d+)/")


def handle_result(timestamp: float, repetitions: int = 1, *args, **kwargs) -> TimeDefined:
    return {"timestamp": timestamp, "repetitions": repetitions}


def handle_timedate(defined: str, start: Optional[datetime.datetime] = None, *args, **kwargs) -> TimeDefined:
    return handle_result(timestamp=isodate.parse_datetime(defined).timestamp())


def handle_time_duration(defined: str, start: Optional[datetime.datetime] = None, *args, **kwargs) -> TimeDefined:
    start = start or datetime.datetime.now()
    return handle_result(timestamp=(start + isodate.parse_duration(defined)).timestamp())


def handle_time_cycle(defined: str, start: Optional[datetime.datetime] = None, *args, **kwargs) -> TimeDefined:
    repeat_match = TIME_CYCLE_DEFINED_PATTERN.match(defined)
    if repeat_match:
        repetitions: int = int(repeat_match.group(1)[1:])
        duration_string = TIME_CYCLE_DEFINED_PATTERN.sub("", defined)
    else:
        repetitions: int = 1
        duration_string = defined

    return handle_result(timestamp=handle_time_duration(duration_string, start)["timestamp"], repetitions=repetitions)


TIMER_TYPE_ROUTES: Dict[str, Callable[[str, Optional[datetime.datetime], Any, Any], TimeDefined]] = {
    TimerType.TIME_DURATION.value: handle_time_duration,
    TimerType.TIME_CYCLE.value: handle_time_cycle,
    TimerType.TIME_DATE.value: handle_timedate,
}


def parse_timer_defined(
    timer_type: str, defined: str, start: Optional[datetime.datetime] = None, *args, **kwargs
) -> TimeDefined:
    if timer_type not in TIMER_TYPE_ROUTES:
        raise ValueError(f"Unsupported timer_type -> {timer_type}")

    return TIMER_TYPE_ROUTES[timer_type](defined, start, *args, **kwargs)



# -*- coding: utf-8 -*-
import json
import logging

from django.apps import apps

from bamboo_engine.builder.builder import generate_pipeline_token

logger = logging.getLogger("bamboo_engine")


class RollbackMixin:
    def set_pipeline_token(self, pipeline_tree: dict):
        """
        设置pipeline token
        """
        try:
            # 引用成功说明pipeline rollback 这个 app 是安装过的
            RollbackToken = apps.get_model("rollback", "RollbackToken")
        except Exception as e:
            logger.error(
                "[RollbackMixin][set_pipeline_token] import RollbackToken error, "
                "Please check whether the  rollback app is installed correctly, err={}".format(e)
            )
            return

        root_pipeline_id = pipeline_tree["id"]
        node_map = generate_pipeline_token(pipeline_tree)

        RollbackToken.objects.create(root_pipeline_id=root_pipeline_id, token=json.dumps(node_map))

    def set_node_snapshot(self, root_pipeline_id, node_id, code, version, context_values, inputs, outputs):
        """
        创建一份节点快照
        """
        try:
            RollbackNodeSnapshot = apps.get_model("rollback", "RollbackNodeSnapshot")
            # 引用成功说明pipeline rollback 这个 app 是安装过的
        except Exception as e:
            logger.error(
                "[RollbackMixin][set_node_snapshot] import RollbackNodeSnapshot error, "
                "Please check whether the rollback app is installed correctly, err={}".format(e)
            )
            return

        RollbackNodeSnapshot.objects.create(
            root_pipeline_id=root_pipeline_id,
            node_id=node_id,
            code=code,
            version=version,
            context_values=context_values,
            inputs=inputs,
            outputs=outputs,
        )

    def start_rollback(self, root_pipeline_id, node_id):
        """
        新建一个回滚任务
        """
        try:
            # 引用成功说明pipeline rollback 这个 app 是安装过的
            from pipeline.contrib.rollback.handler import RollbackDispatcher

            RollbackPlan = apps.get_model("rollback", "RollbackPlan")
        except Exception as e:
            logger.error(
                "[RollbackMixin][set_pipeline_token] import RollbackDispatcher or RollbackPlan error, "
                "Please check whether the rollback app is installed correctly, err={}".format(e)
            )
            return

        try:
            rollback_plan = RollbackPlan.objects.get(
                root_pipeline_id=root_pipeline_id, start_node_id=node_id, is_expired=False
            )
            handler = RollbackDispatcher(root_pipeline_id=root_pipeline_id, mode=rollback_plan.mode)
            handler.rollback(rollback_plan.start_node_id, rollback_plan.target_node_id, **rollback_plan.options)
            rollback_plan.is_expired = True
            rollback_plan.save(update_fields=["is_expired"])
        except Exception as e:
            logger.error("[RollbackMixin][start_rollback] start a rollback task error, err={}".format(e))
            return



# Generated by Django 2.2.8 on 2020-06-10 06:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0025_multicallbackdata"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pipelineprocess",
            name="is_alive",
            field=models.BooleanField(db_index=True, default=True, verbose_name="该 process 是否还有效"),
        ),
        migrations.AlterField(
            model_name="pipelineprocess",
            name="is_frozen",
            field=models.BooleanField(db_index=True, default=False, verbose_name="该 process 是否被冻结"),
        ),
        migrations.AlterField(
            model_name="pipelineprocess",
            name="is_sleep",
            field=models.BooleanField(db_index=True, default=False, verbose_name="该 process 是否正在休眠"),
        ),
        migrations.AlterField(
            model_name="pipelineprocess",
            name="root_pipeline_id",
            field=models.CharField(db_index=True, max_length=32, verbose_name="根 pipeline 的 ID"),
        ),
        migrations.AlterField(
            model_name="scheduleservice",
            name="is_scheduling",
            field=models.BooleanField(db_index=True, default=False, verbose_name="是否正在被调度"),
        ),
        migrations.AlterField(
            model_name="noderelationship",
            name="distance",
            field=models.IntegerField(db_index=True, verbose_name="距离"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import datetime
import logging

from django.dispatch import receiver

from bamboo_engine import states as bamboo_engine_states
from pipeline.contrib.node_timeout.models import TimeoutNodeConfig
from pipeline.contrib.node_timeout.settings import node_timeout_settings
from pipeline.eri.signals import post_set_state


logger = logging.getLogger(__name__)


def _node_timeout_info_update(redis_inst, to_state, node_id, version):
    key = f"{node_id}_{version}"
    if to_state == bamboo_engine_states.RUNNING:
        now = datetime.datetime.now()
        timeout_node_config = TimeoutNodeConfig.objects.filter(node_id=node_id).only("timeout").first()
        if not timeout_node_config:
            return
        timeout_time = (now + datetime.timedelta(seconds=timeout_node_config.timeout)).timestamp()
        redis_inst.zadd(node_timeout_settings.executing_pool, mapping={key: timeout_time}, nx=True)
    elif to_state in [bamboo_engine_states.FAILED, bamboo_engine_states.FINISHED, bamboo_engine_states.SUSPENDED]:
        redis_inst.zrem(node_timeout_settings.executing_pool, key)


@receiver(post_set_state)
def bamboo_engine_eri_node_state_handler(sender, node_id, to_state, version, root_id, parent_id, loop, **kwargs):
    try:
        _node_timeout_info_update(node_timeout_settings.redis_inst, to_state, node_id, version)
    except Exception:
        logger.exception("node_timeout_info_update error")



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.apps import AppConfig
from django.conf import settings
from pipeline.exceptions import ConfigValidationError

from bamboo_engine.handlers import register
from bamboo_engine.utils.constants import ExclusiveGatewayStrategy


class ERIConfig(AppConfig):
    name = "pipeline.eri"
    verbose_name = "PipelineEngineRuntimeInterface"

    def ready(self):
        from .celery.tasks import execute, schedule  # noqa

        register()

        # 校验 PIPELINE_EXCLUSIVE_GATEWAY_EXPR_FUNC 配置
        if hasattr(settings, "PIPELINE_EXCLUSIVE_GATEWAY_EXPR_FUNC"):

            pipeline_exclusive_gateway_expr_func = getattr(settings, "PIPELINE_EXCLUSIVE_GATEWAY_EXPR_FUNC", None)
            if not callable(pipeline_exclusive_gateway_expr_func):
                raise ConfigValidationError("config validate error, the expr func must be callable, please check it")

            # 是否校验
            pipeline_exclusive_gateway_expr_func_check = getattr(
                settings, "PIPELINE_EXCLUSIVE_GATEWAY_EXPR_FUNC_CHECK", True
            )
            if pipeline_exclusive_gateway_expr_func_check:
                # 获取校验文本
                pipeline_exclusive_gateway_expr_func_text = getattr(
                    settings, "PIPELINE_EXCLUSIVE_GATEWAY_EXPR_FUNC_TEXT", "1==1"
                )
                check_result = pipeline_exclusive_gateway_expr_func(pipeline_exclusive_gateway_expr_func_text, {}, {})
                if not check_result:
                    raise ConfigValidationError("config validate error, the expr func return False")

        if hasattr(settings, "PIPELINE_EXCLUSIVE_GATEWAY_STRATEGY"):
            pipeline_exclusive_gateway_strategy = getattr(settings, "PIPELINE_EXCLUSIVE_GATEWAY_STRATEGY")
            if pipeline_exclusive_gateway_strategy not in [
                ExclusiveGatewayStrategy.ONLY.value,
                ExclusiveGatewayStrategy.FIRST.value,
            ]:
                raise ConfigValidationError(
                    "config validate error, the pipeline_exclusive_gateway_strategy only support 1(ONLY), 2(FIRST)"
                )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from enum import Enum


class TimerType(Enum):

    # 时间日期（Time date）ISO 8601 组合日期和时间格式
    # 2019-10-01T12:00:00Z - UTC 时间
    # 2019-10-02T08:09:40+02:00 - UTC 加上两小时时区偏移
    # 2019-10-02T08:09:40+02:00[Europe/Berlin] - UTC 加上柏林两小时时区偏移
    TIME_DATE = "time_date"

    # 时间周期（Time cycle）ISO 8601 重复间隔格式，包含重复次数模式：R(n) 及持续时间模式：P(n)Y(n)M(n)DT(n)H(n)M(n)S
    # R5/PT10S - 每10秒一次，最多五次
    # R/P1D - 每天，无限
    TIME_CYCLE = "time_cycle"

    # 持续时间（Time duration）ISO 8601 持续时间格式，模式：P(n)Y(n)M(n)DT(n)H(n)M(n)S
    # PT15S - 15 秒
    # PT1H30M - 1 小时 30 分钟
    # P14D - 14 天
    # P14DT1H30M - 14 天 1 小时 30 分钟
    TIME_DURATION = "time_duration"



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from importlib import import_module

from django.conf import settings
from pipeline.contrib.node_timer_event import types


def get_import_path(cls: types.T) -> str:
    return f"{cls.__module__}.{cls.__name__}"


def import_string(dotted_path: str):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute/class') from err


class NodeTimerEventSettngs:
    PREFIX = "PIPELINE_NODE_TIMER_EVENT"
    DEFAULT_SETTINGS = {
        # # Redis key 前缀，用于记录正在执行的节点，命名示例: {app_code}:{app_env}:{module}:node_timer_event
        # v1 表示 node_timer_event 的版本，预留以隔离
        "key_prefix": "bamboo:v1:node_timer_event",
        # 节点计时器边界事件处理队列名称, 用于处理计时器边界事件， 需要 worker 接收该队列消息，默认为 None，即使用 celery 默认队列
        "dispatch_queue": None,
        # 节点计时器边界事件分发队列名称, 用于记录计时器边界事件， 需要 worker 接收该队列消息，默认为 None，即使用 celery 默认队列
        "handle_queue": None,
        # 执行节点池名称，用于记录正在执行的节点，需要保证 Redis key 唯一，命名示例: {app_code}:{app_env}:{module}:executing_node_pool
        "executing_pool": "bamboo:v1:node_timer_event:executing_node_pool",
        # 节点池扫描间隔，间隔越小，边界事件触发时间越精准，相应的事件处理的 workload 负载也会提升
        "pool_scan_interval": 1,
        # 最长过期时间，兜底删除 Redis 冗余数据，默认为 15 Days，请根据业务场景调整
        "max_expire_time": 60 * 60 * 24 * 15,
        # 边界事件处理适配器，默认为 `pipeline.contrib.node_timer_event.adapter.NodeTimerEventAdapter`
        "adapter_class": "pipeline.contrib.node_timer_event.adapter.NodeTimerEventAdapter",
    }

    def __getattr__(self, item: str):
        if item == "redis_inst":
            return settings.redis_inst
        if item == "adapter_class":
            return import_string(getattr(settings, f"{self.PREFIX}_{item.upper()}", self.DEFAULT_SETTINGS.get(item)))

        return getattr(settings, f"{self.PREFIX}_{item.upper()}", self.DEFAULT_SETTINGS.get(item))


node_timer_event_settings = NodeTimerEventSettngs()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
import logging
from typing import Any, Dict, List, Type, Union

from celery import current_app
from pipeline.contrib.node_timer_event.adapter import NodeTimerEventBaseAdapter
from pipeline.contrib.node_timer_event.handlers import ActionManager
from pipeline.contrib.node_timer_event.models import ExpiredNodesRecord
from pipeline.contrib.node_timer_event.settings import node_timer_event_settings
from pipeline.eri.models import Process, State

logger = logging.getLogger("celery")


@current_app.task(acks_late=True)
def dispatch_expired_nodes(record_id: int):
    record: ExpiredNodesRecord = ExpiredNodesRecord.objects.get(id=record_id)
    node_keys: List[str] = json.loads(record.nodes)
    logger.info("[dispatch_expired_nodes] record -> %s, nodes -> %s", record_id, node_keys)

    adapter_class: Type[NodeTimerEventBaseAdapter] = node_timer_event_settings.adapter_class

    for node_key in node_keys:
        try:
            key_info: Dict[str, Union[str, int]] = adapter_class.parse_event_key(node_key)
        except ValueError:
            logger.warning(
                "[dispatch_expired_nodes] failed to parse key, skipped: record -> %s, node_key -> %s",
                record_id,
                node_key,
            )
            continue

        index: int = key_info["index"]
        node_id: str = key_info["node_id"]
        version: str = key_info["version"]

        if node_timer_event_settings.handle_queue is None:
            execute_node_timer_event_action.apply_async(kwargs={"node_id": node_id, "version": version, "index": index})
        else:
            execute_node_timer_event_action.apply_async(
                kwargs={"node_id": node_id, "version": version, "index": index},
                queue=node_timer_event_settings.handle_queue,
                routing_key=node_timer_event_settings.handle_queue,
            )

    logger.info("[dispatch_expired_nodes] dispatch finished: record -> %s, nodes -> %s", record_id, node_keys)
    # 删除临时记录
    record.delete()
    logger.info("[dispatch_expired_nodes] record deleted: record -> %s", record_id)


@current_app.task(ignore_result=True)
def execute_node_timer_event_action(node_id: str, version: str, index: int):

    adapter_class: Type[NodeTimerEventBaseAdapter] = node_timer_event_settings.adapter_class
    adapter: NodeTimerEventBaseAdapter = adapter_class(node_id=node_id, version=version)
    if not adapter.is_ready() or (adapter.index__event_map and index not in adapter.index__event_map):
        message: str = (
            f"[execute_node_timer_event_action] no timer config: "
            f"node_id -> {node_id}, version -> {version}, index -> {index}"
        )
        logger.exception(message)
        return {"result": False, "message": message, "data": None}

    event: Dict[str, Any] = adapter.index__event_map[index]

    # 判断当前节点是否符合策略执行要求
    is_process_current_node: bool = Process.objects.filter(
        root_pipeline_id=adapter.root_pipeline_id, current_node_id=node_id
    ).exists()
    is_node_match = State.objects.filter(node_id=node_id, version=version).exists()
    if not (is_node_match and is_process_current_node):
        message = (
            f"[execute_node_timer_event_action] node {node_id} with version {version} "
            f"in pipeline {adapter.root_pipeline_id} has been passed."
        )
        logger.error(message)
        return {"result": False, "message": message, "data": None}

    # 计算事件下一次触发事件并丢进待调度节点池
    adapter.add_to_pool(node_timer_event_settings.redis_inst, event)

    try:
        is_success: bool = ActionManager.get_action(
            adapter.root_pipeline_id, node_id, version, event["action"]
        ).notify()
        logger.info(
            f"[execute_node_timer_event_action] node {node_id} with version {version} in pipeline "
            f"{adapter.root_pipeline_id} action result is: {is_success}."
        )
        return {"result": is_success, "data": None}
    except Exception as e:
        logger.exception(
            f"[execute_node_timer_event_action] node {node_id} with version {version} in pipeline "
            f"{adapter.root_pipeline_id} error is: {e}."
        )
        return {"result": False, "data": None, "message": str(e)}



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import abc
import datetime
import json
import logging
import re
from typing import Dict, List, Optional, Union

from pipeline.contrib.node_timer_event.models import NodeTimerEventConfig
from pipeline.contrib.node_timer_event.settings import node_timer_event_settings
from pipeline.contrib.node_timer_event.types import TimerEvent, TimerEvents
from pipeline.contrib.node_timer_event.utils import parse_timer_defined

logger = logging.getLogger(__name__)

EVENT_KEY_PATTERN = re.compile(r".*node:(?P<node_id>.+):version:(?P<version>.+):index:(?P<index>\d+)")


class NodeTimerEventBaseAdapter(abc.ABC):

    node_id: str = None
    version: str = None
    root_pipeline_id: Optional[str] = None
    events: Optional[TimerEvents] = None
    index__event_map: Optional[Dict[int, TimerEvent]] = None

    def __init__(self, node_id: str, version: str):
        self.node_id: str = node_id
        self.version: str = version

    def is_ready(self) -> bool:
        """适配器是否就绪"""
        if not self.events:
            return False
        return True

    def fetch_keys_to_be_rem(self) -> List[str]:
        """
        获取需要被移除的事件 Key
        :return:
        """
        return [self.get_event_key(event) for event in self.events]

    def get_event_key(self, event: TimerEvent) -> str:
        """
        获取事件 Key
        :param event:
        :return:
        """

        # zset 没有按字符串匹配模式批量删除 key 的支持，使用 key 的命名采用已检索的信息进行拼接
        # 之前想把 loop 也维护进去，发觉移除操作非常麻烦，故采用 incr 的方式，单独维护每个事件事件的触发次数
        key_prefix: str = f"{node_timer_event_settings.key_prefix}:node:{self.node_id}:version:{self.version}"
        return f"{key_prefix}:index:{event['index']}"

    @classmethod
    def get_next_expired_time(cls, event: TimerEvent, start: Optional[datetime.datetime] = None) -> float:
        """
        获取时间事件下一次过期时间
        :param event: 事件详情
        :param start: 开始时间，默认取 datetime.now()
        :return:
        """
        return parse_timer_defined(event["timer_type"], event["defined"], start=start or datetime.datetime.now())[
            "timestamp"
        ]

    def add_to_pool(self, redis_inst, event: TimerEvent):

        key: str = self.get_event_key(event)

        expired_time: float = self.get_next_expired_time(event)

        # TODO 考虑 incr & zadd 合并，使用 lua 封装成原子操作
        loop: int = int(redis_inst.incr(key, 1))
        redis_inst.expire(key, node_timer_event_settings.max_expire_time)
        if loop > event["repetitions"]:
            logger.info(
                "[add_to_pool] No need to add: node -> %s, version -> %s, loop -> %s, event -> %s",
                self.node_id,
                self.version,
                loop,
                event,
            )
            return

        redis_inst.zadd(node_timer_event_settings.executing_pool, mapping={key: expired_time}, nx=True)

        logger.info(
            "[add_to_pool] add event to redis: "
            "node_id -> %s, version -> %s, event -> %s, key -> %s, expired_time -> %s",
            self.node_id,
            self.version,
            event,
            key,
            expired_time,
        )

    @classmethod
    def parse_event_key(cls, key: str) -> Dict[str, Union[str, int]]:
        match = EVENT_KEY_PATTERN.match(key)
        if match:
            key_info: Dict[str, Union[str, int]] = match.groupdict()
            # to int
            key_info["index"] = int(key_info["index"])

            return key_info

        else:
            raise ValueError(f"invalid key -> {key}")


class NodeTimerEventAdapter(NodeTimerEventBaseAdapter):
    def __init__(self, node_id: str, version: str):
        super().__init__(node_id, version)

        node_timer_event_config: NodeTimerEventConfig = NodeTimerEventConfig.objects.filter(
            node_id=self.node_id
        ).first()

        if not node_timer_event_config:
            return

        self.root_pipeline_id: str = node_timer_event_config.root_pipeline_id
        self.events: TimerEvents = json.loads(node_timer_event_config.events)
        self.index__event_map: Dict[int, TimerEvent] = {event["index"]: event for event in self.events}



# Generated by Django 3.2.15 on 2023-12-07 13:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("component_framework", "0007_auto_20201112_2244"),
    ]

    operations = [
        migrations.AddField(
            model_name="componentmodel",
            name="is_default_version",
            field=models.BooleanField(default=False, verbose_name="是否是默认版本"),
        ),
    ]



# Generated by Django 2.2.28 on 2022-04-22 02:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eri", "0004_state_inner_loop_"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contextvalue",
            name="code",
            field=models.CharField(blank=True, default="", max_length=128, verbose_name="计算型变量类型唯一标志"),
        ),
        migrations.AlterField(
            model_name="logentry",
            name="message",
            field=models.TextField(blank=True, null=True, verbose_name="日志内容"),
        ),
        migrations.AlterField(
            model_name="logentry",
            name="version",
            field=models.CharField(blank=True, default="", max_length=33, verbose_name="状态版本"),
        ),
        migrations.AlterField(
            model_name="process",
            name="destination_id",
            field=models.CharField(blank=True, default="", max_length=33, verbose_name="执行终点 ID"),
        ),
        migrations.AlterField(
            model_name="process",
            name="queue",
            field=models.CharField(blank=True, default="", max_length=128, verbose_name="所属队列"),
        ),
        migrations.AlterField(
            model_name="process",
            name="suspended_by",
            field=models.CharField(
                blank=True, db_index=True, default="", max_length=33, verbose_name="导致进程暂停的节点 ID"
            ),
        ),
        migrations.AlterField(
            model_name="state",
            name="archived_time",
            field=models.DateTimeField(blank=True, null=True, verbose_name="归档时间"),
        ),
        migrations.AlterField(
            model_name="state",
            name="parent_id",
            field=models.CharField(blank=True, db_index=True, default="", max_length=33, verbose_name="父节点 ID"),
        ),
        migrations.AlterField(
            model_name="state",
            name="root_id",
            field=models.CharField(blank=True, db_index=True, default="", max_length=33, verbose_name="根节点 ID"),
        ),
        migrations.AlterField(
            model_name="state",
            name="started_time",
            field=models.DateTimeField(blank=True, null=True, verbose_name="开始时间"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json

from django.db import transaction
from django.db.models import Q
from pipeline.conf.default_settings import ROLLBACK_QUEUE
from pipeline.contrib.exceptions import RollBackException
from pipeline.contrib.rollback import constants
from pipeline.contrib.rollback.constants import ANY, TOKEN
from pipeline.contrib.rollback.graph import RollbackGraphHandler
from pipeline.contrib.rollback.models import (
    RollbackPlan,
    RollbackSnapshot,
    RollbackToken,
)
from pipeline.contrib.rollback.tasks import any_rollback, token_rollback
from pipeline.core.constants import PE
from pipeline.eri.models import Node, Process, State
from pipeline.eri.runtime import BambooDjangoRuntime

from bamboo_engine import states


class RollbackValidator:
    @staticmethod
    def validate_pipeline(root_pipeline_id):
        pipeline_state = State.objects.filter(node_id=root_pipeline_id).first()
        if not pipeline_state:
            raise RollBackException(
                "rollback failed: pipeline state not exist, pipeline_id={}".format(root_pipeline_id)
            )

        if pipeline_state.name not in [states.RUNNING, states.ROLL_BACK_FAILED]:
            raise RollBackException(
                "rollback failed: the task of non-running state is not allowed to roll back, "
                "pipeline_id={}, state={}".format(root_pipeline_id, pipeline_state.name)
            )

    @staticmethod
    def validate_node(node_id, allow_failed=False):
        node = Node.objects.filter(node_id=node_id).first()
        if node is None:
            raise RollBackException("rollback failed: node not exist, node={}".format(node_id))

        node_detail = json.loads(node.detail)
        if node_detail["type"] not in [PE.ServiceActivity, PE.EmptyStartEvent]:
            raise RollBackException("rollback failed: only allows rollback to ServiceActivity type nodes")

        target_node_state = State.objects.filter(node_id=node_id).first()

        if target_node_state is None:
            raise RollBackException("rollback failed: node state not exist, node={}".format(node_id))

        allow_states = [states.FINISHED]
        if allow_failed:
            allow_states = [states.FINISHED, states.FAILED]
        if target_node_state.name not in allow_states:
            raise RollBackException(
                "rollback failed: only allows rollback to finished node, allowed states {}".format(allow_states)
            )

    @staticmethod
    def validate_token(root_pipeline_id, start_node_id, target_node_id):
        try:
            rollback_token = RollbackToken.objects.get(root_pipeline_id=root_pipeline_id)
        except RollbackToken.DoesNotExist:
            raise RollBackException(
                "rollback failed: pipeline token not exist, pipeline_id={}".format(root_pipeline_id)
            )

        tokens = json.loads(rollback_token.token)

        start_node_token = tokens.get(start_node_id)
        target_node_token = tokens.get(target_node_id)

        if start_node_token is None or target_node_token is None:
            raise RollBackException("rollback failed: token not found, pipeline_id={}".format(root_pipeline_id))

        if start_node_token != target_node_token:
            raise RollBackException(
                "rollback failed: start node token must equal target node, pipeline_id={}".format(root_pipeline_id)
            )

    @staticmethod
    def validate_node_state(root_pipeline_id, start_node_id):
        """
        使用token模式下的回滚，相同token的节点不允许有正在运行的节点
        """
        try:
            rollback_token = RollbackToken.objects.get(root_pipeline_id=root_pipeline_id)
        except RollbackToken.DoesNotExist:
            raise RollBackException(
                "rollback failed: pipeline token not exist, pipeline_id={}".format(root_pipeline_id)
            )

        tokens = json.loads(rollback_token.token)
        start_token = tokens.get(start_node_id)
        if start_token is None:
            raise RollBackException("rollback failed: can't find the not token, node_id={}".format(start_node_id))

        node_id_list = []
        for node_id, token in node_id_list:
            if token == start_token:
                node_id_list.append(node_id)

        if State.objects.filter(node_id__in=node_id_list, name=states.RUNNING).exists():
            raise RollBackException(
                "rollback failed: there is currently the same node that the same token is running, node_id={}".format(
                    start_node_id
                )
            )

    @staticmethod
    def validate_start_node_id(root_pipeline_id, start_node_id):
        """
        回滚的开始节点必须是流程的末尾节点
        """
        if not Process.objects.filter(root_pipeline_id=root_pipeline_id, current_node_id=start_node_id).exists():
            raise RollBackException("rollback failed: The node to be rolled back must be the current node!")


class BaseRollbackHandler:
    mode = None

    def __init__(self, root_pipeline_id):
        self.root_pipeline_id = root_pipeline_id
        self.runtime = BambooDjangoRuntime()
        # 检查pipeline 回滚的合法性
        RollbackValidator.validate_pipeline(root_pipeline_id)

    def get_allowed_rollback_node_id_list(self, start_node_id, **options):
        """
        获取允许回滚的节点范围
        规则：token 一致的节点允许回滚
        """
        try:
            rollback_token = RollbackToken.objects.get(root_pipeline_id=self.root_pipeline_id)
        except RollbackToken.DoesNotExist:
            raise RollBackException(
                "rollback failed: pipeline token not exist, pipeline_id={}".format(self.root_pipeline_id)
            )
        node_map = self.get_allowed_rollback_node_map()
        service_activity_node_list = [
            node_id for node_id, node_detail in node_map.items() if node_detail["type"] == PE.ServiceActivity
        ]

        tokens = json.loads(rollback_token.token)
        start_token = tokens.get(start_node_id)
        if not start_token:
            return []

        nodes = []
        for node_id, token in tokens.items():
            if start_token == token and node_id != start_node_id and node_id in service_activity_node_list:
                nodes.append(node_id)

        return nodes

    def get_allowed_rollback_node_map(self):
        # 不需要遍历整颗树，获取到现在已经执行成功和失败节点的所有列表
        finished_node_id_list = (
            State.objects.filter(root_id=self.root_pipeline_id, name__in=[states.FINISHED, states.FAILED])
            .exclude(node_id=self.root_pipeline_id)
            .values_list("node_id", flat=True)
        )
        node_detail_list = Node.objects.filter(node_id__in=finished_node_id_list)
        # 获取node_id 到 node_detail的映射
        return {n.node_id: json.loads(n.detail) for n in node_detail_list}

    def _reserve(self, start_node_id, target_node_id, reserve_rollback=True, **options):
        # 节点预约 需要在 Node 里面 插入 reserve_rollback = True, 为 True的节点执行完将暂停
        RollbackValidator.validate_start_node_id(self.root_pipeline_id, start_node_id)
        RollbackValidator.validate_node(target_node_id)
        node = Node.objects.filter(node_id=start_node_id).first()
        if node is None:
            raise RollBackException("reserve rollback failed, the node is not exists, node_id={}".format(start_node_id))

        state = State.objects.filter(node_id=start_node_id).first()
        if state is None:
            raise RollBackException(
                "reserve rollback failed, the node state is not exists, node_id={}".format(start_node_id)
            )

        # 不在执行中的节点不允许预约
        if state.name != states.RUNNING:
            raise RollBackException(
                "reserve rollback failed, the node state is not Running, current state={},  node_id={}".format(
                    state.name, start_node_id
                )
            )

        with transaction.atomic():
            if reserve_rollback:
                # 一个流程只能同时拥有一个预约任务
                if RollbackPlan.objects.filter(root_pipeline_id=self.root_pipeline_id, is_expired=False).exists():
                    raise RollBackException("reserve rollback failed, there exists another unfinished rollback plan")
                RollbackPlan.objects.create(
                    root_pipeline_id=self.root_pipeline_id,
                    start_node_id=start_node_id,
                    target_node_id=target_node_id,
                    mode=self.mode,
                    options=options,
                )
            else:
                # 取消回滚，删除所有的任务
                RollbackPlan.objects.filter(root_pipeline_id=self.root_pipeline_id, start_node_id=start_node_id).update(
                    is_expired=True
                )

            node_detail = json.loads(node.detail)
            node_detail["reserve_rollback"] = reserve_rollback
            node.detail = json.dumps(node_detail)
            node.save()

    def reserve_rollback(self, start_node_id, target_node_id, **options):
        """
        预约回滚
        """
        RollbackValidator.validate_token(self.root_pipeline_id, start_node_id, target_node_id)
        self._reserve(start_node_id, target_node_id, **options)

    def cancel_reserved_rollback(self, start_node_id, target_node_id):
        """
        取消预约回滚
        """
        self._reserve(start_node_id, target_node_id, reserve_rollback=False)


class AnyRollbackHandler(BaseRollbackHandler):
    mode = ANY

    def get_allowed_rollback_node_id_list(self, start_node_id, **options):
        # 如果开启了token跳过检查这个选项，那么将返回所有运行过的节点作为回滚范围
        if options.get("skip_check_token", False):
            node_map = self.get_allowed_rollback_node_map()
            service_activity_node_list = [
                node_id
                for node_id, node_detail in node_map.items()
                if node_detail["type"] == PE.ServiceActivity and node_id != start_node_id
            ]
            return service_activity_node_list
        return super(AnyRollbackHandler, self).get_allowed_rollback_node_id_list(start_node_id, **options)

    def retry_rollback_failed_node(self, node_id, retry_data):
        """ """
        raise RollBackException("rollback failed: when mode is any, not support retry")

    def reserve_rollback(self, start_node_id, target_node_id, **options):
        """
        预约回滚
        """
        if not options.get("skip_check_token", False):
            RollbackValidator.validate_token(self.root_pipeline_id, start_node_id, target_node_id)
        self._reserve(start_node_id, target_node_id, **options)

    def rollback(self, start_node_id, target_node_id, skip_rollback_nodes=None, **options):
        RollbackValidator.validate_start_node_id(self.root_pipeline_id, start_node_id)
        RollbackValidator.validate_node(start_node_id, allow_failed=True)
        RollbackValidator.validate_node(target_node_id)
        if not options.get("skip_check_token", False):
            RollbackValidator.validate_token(self.root_pipeline_id, start_node_id, target_node_id)
            # 相同token回滚时，不允许同一token路径上有正在运行的节点
            RollbackValidator.validate_node_state(self.root_pipeline_id, start_node_id)

        node_map = self.get_allowed_rollback_node_map()
        rollback_graph = RollbackGraphHandler(node_map=node_map, start_id=start_node_id, target_id=target_node_id)

        graph, other_nodes = rollback_graph.build_rollback_graph()
        node_access_record = {node: 0 for node in graph.nodes}

        rollback_snapshot = RollbackSnapshot.objects.create(
            root_pipeline_id=self.root_pipeline_id,
            graph=json.dumps(graph.as_dict()),
            node_access_record=json.dumps(node_access_record),
            start_node_id=start_node_id,
            target_node_id=target_node_id,
            other_nodes=json.dumps(other_nodes),
            skip_rollback_nodes=json.dumps([]),
        )

        any_rollback.apply_async(
            kwargs={"snapshot_id": rollback_snapshot.id},
            queue=ROLLBACK_QUEUE,
        )


class TokenRollbackHandler(BaseRollbackHandler):
    mode = TOKEN

    def retry_rollback_failed_node(self, node_id, retry_data):
        """
        重试回滚失败的节点
        """
        pipeline_state = State.objects.filter(node_id=self.root_pipeline_id).first()
        if pipeline_state.name != states.ROLL_BACK_FAILED:
            raise RollBackException(
                "rollback failed: only retry the failed pipeline, current_status={}".format(pipeline_state.name)
            )
        node_state = State.objects.filter(node_id=node_id).first()
        if node_state.name != states.ROLL_BACK_FAILED:
            raise RollBackException(
                "rollback failed: only retry the failed node, current_status={}".format(node_state.name)
            )

        # 获取镜像
        try:
            rollback_snapshot = RollbackSnapshot.objects.get(root_pipeline_id=self.root_pipeline_id, is_expired=False)
        except RollbackSnapshot.DoesNotExist:
            raise RollBackException("rollback failed: the rollback snapshot is not exists, please check")
        except RollbackSnapshot.MultipleObjectsReturned:
            raise RollBackException("rollback failed: found multi not expired rollback snapshot, please check")

        # 重置pipeline的状态为回滚中
        self.runtime.set_state(
            node_id=self.root_pipeline_id,
            to_state=states.ROLLING_BACK,
        )

        # 驱动这个任务
        token_rollback.apply_async(
            kwargs={
                "snapshot_id": rollback_snapshot.id,
                "node_id": node_id,
                "retry": True,
                "retry_data": retry_data,
            },
            queue=ROLLBACK_QUEUE,
        )

    def _node_state_is_failed(self, node_id):
        """
        判断该节点是不是失败的状态
        """
        node_state = State.objects.filter(node_id=node_id).first()
        if node_state.name == states.FAILED:
            return True
        return False

    def _get_failed_skip_node_id_list(self, node_id_list):
        failed_skip_node_id_list = State.objects.filter(
            Q(Q(skip=True) | Q(error_ignored=True)) & Q(node_id__in=node_id_list)
        ).values_list("node_id", flat=True)
        return failed_skip_node_id_list

    def rollback(self, start_node_id, target_node_id, skip_rollback_nodes=None, **options):

        if skip_rollback_nodes is None:
            skip_rollback_nodes = []

        # 回滚的开始节点运行失败的情况
        RollbackValidator.validate_node(start_node_id, allow_failed=True)
        RollbackValidator.validate_node(target_node_id)
        # 相同token回滚时，不允许同一token路径上有正在运行的节点
        RollbackValidator.validate_node_state(self.root_pipeline_id, start_node_id)
        RollbackValidator.validate_token(self.root_pipeline_id, start_node_id, target_node_id)

        # 如果开始节点是失败的情况，则跳过该节点的回滚操作
        if self._node_state_is_failed(start_node_id):
            skip_rollback_nodes.append(start_node_id)

        node_map = self.get_allowed_rollback_node_map()
        rollback_graph = RollbackGraphHandler(node_map=node_map, start_id=start_node_id, target_id=target_node_id)

        runtime = BambooDjangoRuntime()

        graph, other_nodes = rollback_graph.build_rollback_graph()
        node_access_record = {node: 0 for node in graph.nodes}

        # 所有失败并跳过的节点不再参与回滚
        failed_skip_node_id_list = self._get_failed_skip_node_id_list(node_map.keys())
        skip_rollback_nodes.extend(list(failed_skip_node_id_list))

        rollback_snapshot = RollbackSnapshot.objects.create(
            root_pipeline_id=self.root_pipeline_id,
            graph=json.dumps(graph.as_dict()),
            node_access_record=json.dumps(node_access_record),
            start_node_id=start_node_id,
            target_node_id=target_node_id,
            other_nodes=json.dumps(other_nodes),
            skip_rollback_nodes=json.dumps(skip_rollback_nodes),
        )

        runtime.set_state(
            node_id=self.root_pipeline_id,
            to_state=states.ROLLING_BACK,
        )
        # 驱动这个任务
        token_rollback.apply_async(
            kwargs={
                "snapshot_id": rollback_snapshot.id,
                "node_id": constants.START_FLAG,
                "retry": False,
                "retry_data": None,
            },
            queue=ROLLBACK_QUEUE,
        )


class RollbackDispatcher:
    def __init__(self, root_pipeline_id, mode):
        if mode == ANY:
            self.handler = AnyRollbackHandler(root_pipeline_id)
        elif mode == TOKEN:
            self.handler = TokenRollbackHandler(root_pipeline_id)
        else:
            raise RollBackException("rollback failed: not support this mode, please check")

    def rollback(self, start_node_id: str, target_node_id: str, skip_rollback_nodes: list = None, **options):
        self.handler.rollback(start_node_id, target_node_id, skip_rollback_nodes, **options)

    def reserve_rollback(self, start_node_id: str, target_node_id: str, **options):
        self.handler.reserve_rollback(start_node_id, target_node_id, **options)

    def retry_rollback_failed_node(self, node_id: str, retry_data: dict = None):
        self.handler.retry_rollback_failed_node(node_id, retry_data)

    def cancel_reserved_rollback(self, start_node_id: str, target_node_id: str):
        self.handler.cancel_reserved_rollback(start_node_id, target_node_id)

    def get_allowed_rollback_node_id_list(self, start_node_id: str, **options):
        return self.handler.get_allowed_rollback_node_id_list(start_node_id, **options)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from pipeline.conf.default_settings import PLUGIN_EXECUTE_QUEUE
from pipeline.contrib.exceptions import PluginExecuteException
from pipeline.contrib.plugin_execute.contants import State
from pipeline.contrib.plugin_execute.models import PluginExecuteTask, get_schedule_lock
from pipeline.contrib.plugin_execute.tasks import execute, schedule

logger = logging.getLogger("celery")


def _retry_once(action: callable):
    try:
        action()
    except Exception:
        try:
            action()
        except Exception as e:
            raise e


class PluginExecuteHandler:
    @classmethod
    def run(cls, component_code: str, version: str, inputs: dict, contexts: dict, runtime_attrs: dict = None):
        if runtime_attrs is None:
            runtime_attrs = {}

        if not (isinstance(inputs, dict) and isinstance(contexts, dict) and isinstance(runtime_attrs, dict)):
            raise PluginExecuteException("[plugin_execute_run] error, the inputs, contexts, runtime_attrs must be dict")
        plugin_execute_task = PluginExecuteTask.objects.create(
            state=State.READY,
            inputs=inputs,
            version=version,
            component_code=component_code,
            contexts=contexts,
            runtime_attrs=runtime_attrs,
        )

        def action():
            # 发送执行任务
            execute.apply_async(
                kwargs={"task_id": plugin_execute_task.id}, queue=PLUGIN_EXECUTE_QUEUE, ignore_result=True
            )
            logger.info(
                "[plugin_execute_run] send execute task, plugin_execute_task_id = {}".format(plugin_execute_task.id)
            )

        try:
            _retry_once(action=action)
        except Exception as e:
            # 如果任务启动出现异常，则删除任务
            plugin_execute_task.delete()
            raise e

        return plugin_execute_task.id

    @classmethod
    def get_state(cls, task_id):
        """
        获取任务状态
        @param task_id:
        @return:
        """
        # 直接抛出异常让上层去捕获
        plugin_execute_task = PluginExecuteTask.objects.get(id=task_id)
        return {
            "task_id": plugin_execute_task.id,
            "state": plugin_execute_task.state,
            "component_code": plugin_execute_task.component_code,
            "version": plugin_execute_task.version,
            "invoke_count": plugin_execute_task.invoke_count,
            "inputs": plugin_execute_task.inputs,
            "outputs": plugin_execute_task.outputs,
            "callback_data": plugin_execute_task.callback_data,
            "contexts": plugin_execute_task.contexts,
            "runtime_attrs": plugin_execute_task.runtime_attrs,
            "create_at": plugin_execute_task.created_at,
            "finish_at": plugin_execute_task.finish_at,
        }

    @classmethod
    def forced_fail(cls, task_id):
        plugin_execute_task = PluginExecuteTask.objects.get(id=task_id)
        if plugin_execute_task.state != State.RUNNING:
            raise PluginExecuteException(
                "[forced_fail] error, the plugin_execute_task.state is not RUNNING, state={}".format(
                    plugin_execute_task.state
                )
            )
        # 插件状态改成 FAILED, 在schdule会自动停止
        plugin_execute_task.state = State.FAILED
        plugin_execute_task.save()

    @classmethod
    def callback(cls, task_id: int, callback_data: dict = None):

        if callback_data is None:
            callback_data = {}

        if not isinstance(callback_data, dict):
            raise PluginExecuteException("[plugin_execute_callback] error, the callback must be dict")

        plugin_execute_task = PluginExecuteTask.objects.get(id=task_id)
        if plugin_execute_task.state != State.RUNNING:
            raise PluginExecuteException(
                "[callback] error, the plugin_execute_task.state is not RUNNING, state={}".format(
                    plugin_execute_task.state
                )
            )

        def action():
            # 需要加锁，防止流程处在回调的过程中
            with get_schedule_lock(task_id) as locked:
                if not locked:
                    raise PluginExecuteException("[plugin_execute_callback] error, it`s have callback task, please try")
                plugin_execute_task.callback_data = callback_data
                plugin_execute_task.save()
                schedule.apply_async(
                    kwargs={"task_id": plugin_execute_task.id}, queue=PLUGIN_EXECUTE_QUEUE, ignore_result=True
                )
                logger.info("[plugin_execute_callback] send callback task, plugin_execute_task_id = {}".format(task_id))

        _retry_once(action=action)

    @classmethod
    def retry_node(cls, task_id: int, inputs: dict = None, contexts: dict = None, runtime_attrs: dict = None):

        plugin_execute_task = PluginExecuteTask.objects.get(id=task_id)
        if plugin_execute_task.state != State.FAILED:
            raise PluginExecuteException(
                "[retry_node] error, the plugin_execute_task.state is not FAILED, state={}".format(
                    plugin_execute_task.state
                )
            )

        if contexts and isinstance(contexts, dict):
            plugin_execute_task.contexts = contexts
        if inputs and isinstance(inputs, dict):
            plugin_execute_task.inputs = inputs
        if runtime_attrs and isinstance(runtime_attrs, dict):
            plugin_execute_task.runtime_attrs = runtime_attrs

        plugin_execute_task.state = State.READY
        plugin_execute_task.inputs = inputs
        plugin_execute_task.invoke_count += 1
        # 清空输出和callback_data
        plugin_execute_task.outputs = {}
        plugin_execute_task.callback_data = {}
        plugin_execute_task.save()

        def action():
            execute.apply_async(kwargs={"task_id": plugin_execute_task.id}, queue=PLUGIN_EXECUTE_QUEUE)
            logger.info(
                "[plugin_execute_retry_node] send retry_node task,  plugin_execute_task_id = {}".format(task_id)
            )

        _retry_once(action=action)



# Generated by Django 3.2.18 on 2023-11-28 11:49

import pipeline.contrib.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="PluginExecuteTask",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("state", models.CharField(max_length=64, verbose_name="状态名")),
                ("invoke_count", models.IntegerField(default=1, verbose_name="invoke count")),
                ("component_code", models.CharField(db_index=True, max_length=255, verbose_name="组件编码")),
                ("version", models.CharField(default="legacy", max_length=255, verbose_name="插件版本")),
                ("inputs", pipeline.contrib.fields.SerializerField(default={}, verbose_name="node inputs")),
                ("outputs", pipeline.contrib.fields.SerializerField(default={}, verbose_name="node outputs")),
                ("callback_data", pipeline.contrib.fields.SerializerField(default={}, verbose_name="callback data")),
                (
                    "contexts",
                    pipeline.contrib.fields.SerializerField(default={}, verbose_name="pipeline context values"),
                ),
                ("runtime_attrs", pipeline.contrib.fields.SerializerField(default={}, verbose_name="runtime attr")),
                ("scheduling", models.BooleanField(db_index=True, default=False, verbose_name="是否正在调度")),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="create time")),
                ("finish_at", models.DateTimeField(null=True, verbose_name="finish time")),
            ],
        ),
        migrations.AddIndex(
            model_name="pluginexecutetask",
            index=models.Index(fields=["id", "scheduling"], name="plugin_exec_id_e7c7b2_idx"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import copy

from pipeline.contrib.node_timeout.models import TimeoutNodeConfig
from pipeline.core.constants import PE


def apply_node_timout_configs(pipeline_tree: dict, configs: dict):
    """
    在 pipeline_tree 中应用节点超时配置
    :param pipeline_tree: pipeline_tree
    :param configs: 节点超时配置, e.g. {"node_id": {"enable": True, "action": "forced_fail", "seconds": "10"}}
    """
    new_pipeline_tree = copy.deepcopy(pipeline_tree)
    for act_id, act in pipeline_tree[PE.activities].items():
        if act["type"] == PE.SubProcess:
            apply_node_timout_configs(act[PE.pipeline], configs)
        elif act["type"] == PE.ServiceActivity and act_id in configs:
            act["timeout_config"] = {
                "enable": configs[act_id]["enable"],
                "action": configs[act_id]["action"],
                "seconds": configs[act_id]["seconds"],
            }
    return new_pipeline_tree


def batch_create_node_timeout_config(root_pipeline_id: str, pipeline_tree: dict):
    """
    批量创建节点超时配置
    :param root_pipeline_id: pipeline root ID
    :param pipeline_tree: pipeline_tree
    :return: 节点超时配置数据插入结果，e.g. {"result": True, "data": objs, "message": ""}
    """
    return TimeoutNodeConfig.objects.batch_create_node_timeout_config(root_pipeline_id, pipeline_tree)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
import logging

from celery import current_app

from pipeline.contrib.node_timeout.settings import node_timeout_settings
from pipeline.eri.models import State, Process

from pipeline.contrib.node_timeout.models import TimeoutNodeConfig, TimeoutNodesRecord


logger = logging.getLogger("celery")


@current_app.task(acks_late=True)
def dispatch_timeout_nodes(record_id: int):
    record = TimeoutNodesRecord.objects.get(id=record_id)
    nodes = json.loads(record.timeout_nodes)
    for node in nodes:
        node_id, version = node.split("_")
        if node_timeout_settings.handle_queue is None:
            execute_node_timeout_strategy.apply_async(kwargs={"node_id": node_id, "version": version})
        else:
            execute_node_timeout_strategy.apply_async(
                kwargs={"node_id": node_id, "version": version},
                queue=node_timeout_settings.handle_queue,
                routing_key=node_timeout_settings.handle_queue,
            )


@current_app.task(ignore_result=True)
def execute_node_timeout_strategy(node_id, version):
    timeout_config = TimeoutNodeConfig.objects.filter(node_id=node_id).only("root_pipeline_id", "action").first()
    if timeout_config is None:
        logger.error(f"[execute_node_timeout_strategy] node {node_id} with version {version} has no timeout config.")
        return {
            "result": False,
            "message": f"no timeout config with node {node_id} and version {version}",
            "data": None,
        }

    root_pipeline_id, action = timeout_config.root_pipeline_id, timeout_config.action

    # 判断当前节点是否符合策略执行要求
    is_process_current_node = Process.objects.filter(
        root_pipeline_id=root_pipeline_id, current_node_id=node_id
    ).exists()
    node_match = State.objects.filter(node_id=node_id, version=version).exists()
    if not (node_match and is_process_current_node):
        message = (
            f"[execute_node_timeout_strategy] node {node_id} with version {version} "
            f"in pipeline {root_pipeline_id} has been passed."
        )
        logger.error(message)
        return {"result": False, "message": message, "data": None}

    try:
        handler = node_timeout_settings.handler.get(action)
        action_result = handler.deal_with_timeout_node(node_id)
        logger.info(
            f"[execute_node_timeout_strategy] node {node_id} with version {version} in pipeline {root_pipeline_id} "
            f"action result is: {action_result}."
        )
    except Exception as e:
        logger.exception(
            f"[execute_node_timeout_strategy] node {node_id} with version {version} in pipeline {root_pipeline_id} "
            f"error is: {e}."
        )
        return {"result": False, "data": None, "message": str(e)}

    return action_result



# Generated by Django 2.2.24 on 2022-01-12 14:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0031_auto_20210624_2317"),
    ]

    operations = [
        migrations.AddField(
            model_name="templatescheme",
            name="subprocess_scheme_relation",
            field=models.ManyToManyField(
                to="pipeline.TemplateRelationship", verbose_name="子流程节点引用执行方案的关系"
            ),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json

from django.http import JsonResponse
from django.test import TestCase

from bamboo_engine.api import EngineAPIResult
from pipeline.contrib.engine_admin.views import _ensure_return_json_response, _check_api_permission
from pipeline.engine.utils import ActionResult


def check_permission_fail(request, *args, **kwargs):
    return False


def check_permission_success(request, *args, **kwargs):
    return True


class EngineAdminTestCase(TestCase):

    def test_ensure_return_json_response(self):
        def func_return_json_response(*args, **kwargs):
            return JsonResponse({"result": True})

        def func_return_engine_api_result(*args, **kwargs):
            return EngineAPIResult(result=True, message="success")

        def func_return_action_result(*args, **kwargs):
            return ActionResult(result=True, message="success")

        self.assertIsInstance(_ensure_return_json_response(func_return_json_response)(), JsonResponse)
        self.assertIsInstance(_ensure_return_json_response(func_return_engine_api_result)(), JsonResponse)
        self.assertIsInstance(_ensure_return_json_response(func_return_action_result)(), JsonResponse)

    def test_check_api_permission(self):
        def func_return_json_response(request, *args, **kwargs):
            return JsonResponse({"result": True})

        with self.settings(
            PIPELINE_ENGINE_ADMIN_API_PERMISSION="pipeline.contrib.engine_admin.tests.check_permission_fail"
        ):
            fail_response = _check_api_permission(func_return_json_response)(request=None)
            fail_result = json.loads(fail_response.content).get("result")
            self.assertFalse(fail_result)

        with self.settings(
            PIPELINE_ENGINE_ADMIN_API_PERMISSION="pipeline.contrib.engine_admin.tests.check_permission_success"
        ):
            success_response = _check_api_permission(func_return_json_response)(request=None)
            success_result = json.loads(success_response.content).get("result")
            self.assertTrue(success_result)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback

from celery import current_app
from django.utils import timezone
from pipeline.component_framework.library import ComponentLibrary
from pipeline.conf.default_settings import PLUGIN_EXECUTE_QUEUE
from pipeline.contrib.plugin_execute.contants import State
from pipeline.contrib.plugin_execute.models import PluginExecuteTask
from pipeline.core.data.base import DataObject

logger = logging.getLogger("celery")


@current_app.task
def execute(task_id):
    try:
        plugin_execute_task = PluginExecuteTask.objects.get(id=task_id)
    except PluginExecuteTask.DoesNotExist:
        logger.exception("[plugin_execute] execute error, task not exist, task_id={}".format(task_id))
        return

    # 更新插件的状态
    plugin_execute_task.state = State.RUNNING
    plugin_execute_task.save(update_fields=["state"])

    # 封装data
    data = DataObject(inputs=plugin_execute_task.inputs, outputs={})
    parent_data = DataObject(inputs=plugin_execute_task.contexts, outputs={})

    try:
        # 获取 component
        comp_cls = ComponentLibrary.get_component_class(plugin_execute_task.component_code, plugin_execute_task.version)
        # 获取service
        service = comp_cls.bound_service(name=plugin_execute_task.runtime_attrs.get("name", None))

        # 封装运行时
        service.setup_runtime_attrs(**plugin_execute_task.runtime_attrs, logger=logger)
        execute_success = service.execute(data, parent_data)
        # 在 pipeline 中，如果插件返回为None,则表示成功
        if execute_success is None:
            execute_success = True
        plugin_execute_task.outputs = data.outputs
        plugin_execute_task.save()
    except Exception as e:
        # 处理异常情况
        ex_data = traceback.format_exc()
        data.outputs.ex_data = ex_data
        logger.exception("[plugin_execute] plugin execute failed, err={}".format(e))
        plugin_execute_task.outputs = data.outputs
        plugin_execute_task.state = State.FAILED
        plugin_execute_task.save()
        return

    # 单纯的执行失败, 更新状态和输出信息
    if not execute_success:
        plugin_execute_task.state = State.FAILED
        plugin_execute_task.save()
        return

    # 执行成功, 需要判断是否需要调度
    need_schedule = service.need_schedule()
    if not need_schedule:
        plugin_execute_task.state = State.FINISHED
        plugin_execute_task.finish_at = timezone.now()
        plugin_execute_task.save()
        return

    # 需要调度，则调度自身
    if service.interval:
        schedule.apply_async(
            kwargs={"task_id": task_id},
            queue=PLUGIN_EXECUTE_QUEUE,
            countdown=service.interval.next(),
            ignore_result=True,
        )


@current_app.task
def schedule(task_id):
    try:
        plugin_execute_task = PluginExecuteTask.objects.get(id=task_id)
    except PluginExecuteTask.DoesNotExist:
        logger.exception("[plugin_execute] schedule error, task not exist, task_id={}".format(task_id))
        return

    # 只有处于运行状态的节点才允许被调度
    if plugin_execute_task.state != State.RUNNING:
        logger.exception("[plugin_execute] schedule error, task not exist, task_id={}".format(task_id))
        return

    data = DataObject(inputs=plugin_execute_task.inputs, outputs=plugin_execute_task.outputs)
    parent_data = DataObject(inputs=plugin_execute_task.contexts, outputs={})

    try:
        comp_cls = ComponentLibrary.get_component_class(plugin_execute_task.component_code, plugin_execute_task.version)
        # 获取service
        service = comp_cls.bound_service(name=plugin_execute_task.runtime_attrs.get("name", None))
        # 封装运行时
        service.setup_runtime_attrs(**plugin_execute_task.runtime_attrs, logger=logger)
        schedule_success = service.schedule(
            data=data, parent_data=parent_data, callback_data=plugin_execute_task.callback_data
        )
        # 在 pipeline 中，如果插件返回为None,则表示成功
        if schedule_success is None:
            schedule_success = True
        plugin_execute_task.outputs = data.outputs
        plugin_execute_task.save()
    except Exception as e:
        # 处理异常情况
        ex_data = traceback.format_exc()
        data.outputs.ex_data = ex_data
        logger.exception("[plugin_execute] plugin execute failed, err={}".format(e))
        plugin_execute_task.outputs = data.outputs
        plugin_execute_task.state = State.FAILED
        plugin_execute_task.save()
        return

    if not schedule_success:
        plugin_execute_task.state = State.FAILED
        plugin_execute_task.save()
        return

    if service.is_schedule_finished():
        plugin_execute_task.state = State.FINISHED
        plugin_execute_task.finish_at = timezone.now()
        plugin_execute_task.save()
        return

    # 还需要下一次的调度
    # 需要调度，则调度自身
    if service.interval:
        schedule.apply_async(
            kwargs={"task_id": task_id},
            queue=PLUGIN_EXECUTE_QUEUE,
            countdown=service.interval.next(),
            ignore_result=True,
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import logging

from django.db import models
from django.utils.translation import gettext_lazy as _
from pipeline.core.constants import PE

logger = logging.getLogger(__name__)


class TimeoutNodeConfigManager(models.Manager):
    def parse_node_timeout_configs(self, pipeline_tree: dict) -> list:
        configs = []
        for act_id, act in pipeline_tree[PE.activities].items():
            if act["type"] == PE.SubProcess:
                result = self.parse_node_timeout_configs(act[PE.pipeline])
                if not result["result"]:
                    return result
                configs.extend(result["data"])
            elif act["type"] == PE.ServiceActivity:
                timeout_config = act.get("timeout_config", {})
                enable = timeout_config.get("enable")
                if not enable:
                    continue
                timeout_seconds = timeout_config.get("seconds")
                action = timeout_config.get("action")
                if not timeout_seconds or not isinstance(timeout_seconds, int):
                    logger.error(
                        f"[parse_node_timeout_configs] node [{act_id}] timeout "
                        f"config seconds [{timeout_seconds}] is invalid"
                    )
                    # 对于不符合格式要求的情况，则不设置对应超时时间
                    continue
                configs.append({"action": action, "node_id": act_id, "timeout": timeout_seconds})
        return {"result": True, "data": configs, "message": ""}

    def batch_create_node_timeout_config(self, root_pipeline_id: str, pipeline_tree: dict):
        """批量创建节点超时配置"""

        config_parse_result = self.parse_node_timeout_configs(pipeline_tree)
        # 这里忽略解析失败的情况，保证即使解析失败也能正常创建任务
        if not config_parse_result["result"]:
            logger.error(
                f'[batch_create_node_timeout_config] parse node timeout config failed: {config_parse_result["result"]}'
            )
            return config_parse_result
        configs = config_parse_result["data"] or []
        config_objs = [
            TimeoutNodeConfig(
                action=config["action"],
                root_pipeline_id=root_pipeline_id,
                node_id=config["node_id"],
                timeout=config["timeout"],
            )
            for config in configs
        ]
        objs = self.bulk_create(config_objs, batch_size=1000)
        return {"result": True, "data": objs, "message": ""}


class TimeoutNodeConfig(models.Model):
    root_pipeline_id = models.CharField(verbose_name="root pipeline id", max_length=64)
    action = models.CharField(verbose_name="action", max_length=32)
    node_id = models.CharField(verbose_name="task node id", max_length=64, primary_key=True)
    timeout = models.IntegerField(verbose_name="node timeout time")

    objects = TimeoutNodeConfigManager()

    class Meta:
        verbose_name = _("节点超时配置 TimeoutNodeConfig")
        verbose_name_plural = _("节点超时配置 TimeoutNodeConfig")
        index_together = [("root_pipeline_id", "node_id")]


class TimeoutNodesRecord(models.Model):
    id = models.BigAutoField(verbose_name="ID", primary_key=True)
    timeout_nodes = models.TextField(verbose_name="超时节点信息")

    class Meta:
        verbose_name = _("超时节点数据记录 TimeoutNodesRecord")
        verbose_name_plural = _("超时节点数据记录 TimeoutNodesRecord")



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import datetime
import json
import logging
import signal
import time

from django.core.management import BaseCommand
from django.db import connections
from pipeline.contrib.node_timer_event.models import ExpiredNodesRecord
from pipeline.contrib.node_timer_event.settings import node_timer_event_settings
from pipeline.contrib.node_timer_event.tasks import dispatch_expired_nodes

logger = logging.getLogger("root")


class Command(BaseCommand):
    help = "scanning expired nodes and dispatch them to celery task"
    has_killed = False

    def handle(self, *args, **options):
        signal.signal(signal.SIGTERM, self._graceful_exit)
        redis_inst = node_timer_event_settings.redis_inst
        nodes_pool = node_timer_event_settings.executing_pool
        while not self.has_killed:
            try:
                start = time.time()
                self._pop_expired_nodes(redis_inst, nodes_pool)
                end = time.time()
                logger.info(f"[node_timeout_process] time consuming: {end-start}")
            except Exception as e:
                logger.exception(e)
            time.sleep(node_timer_event_settings.pool_scan_interval)

    def _graceful_exit(self, *args):
        self.has_killed = True

    def _pop_expired_nodes(self, redis_inst, nodes_pool) -> list:
        now = datetime.datetime.now().timestamp()
        expired_nodes = [
            node.decode("utf-8") if isinstance(node, bytes) else node
            for node in redis_inst.zrangebyscore(nodes_pool, "-inf", now)
        ]
        if expired_nodes:
            self.record_expired_nodes(expired_nodes)
            redis_inst.zrem(nodes_pool, *expired_nodes)
        return expired_nodes

    @staticmethod
    def record_expired_nodes(timeout_nodes: list):
        # 处理因为过长时间没有访问导致的链接失效问题
        for conn in connections.all():
            conn.close_if_unusable_or_obsolete()

        record = ExpiredNodesRecord.objects.create(nodes=json.dumps(timeout_nodes))
        if node_timer_event_settings.dispatch_queue is None:
            dispatch_expired_nodes.apply_async(kwargs={"record_id": record.id})
        else:
            dispatch_expired_nodes.apply_async(
                kwargs={"record_id": record.id},
                queue=node_timer_event_settings.dispatch_queue,
                routing_key=node_timer_event_settings.dispatch_queue,
            )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import logging
from abc import abstractmethod
from typing import Any, Dict, Tuple

from pipeline.core.data.base import DataObject
from pipeline.eri.runtime import BambooDjangoRuntime

from bamboo_engine import api as bamboo_engine_api
from bamboo_engine.eri import ExecutionData

logger = logging.getLogger(__name__)


class ActionManager:
    __hub = {}

    @classmethod
    def register_invocation_cls(cls, invocation_cls):
        action_name = invocation_cls.Meta.action_name
        existed_invocation_cls = cls.__hub.get(action_name)
        if existed_invocation_cls:
            raise RuntimeError(
                "func register error, {}'s action_name {} conflict with {}".format(
                    existed_invocation_cls, action_name, invocation_cls
                )
            )

        cls.__hub[action_name] = invocation_cls

    @classmethod
    def clear(cls):
        cls.__hub = {}

    @classmethod
    def get_action(cls, root_pipeline_id: str, node_id: str, version: str, action_name: str) -> "BaseAction":
        """
        获取 Action 实例
        :param root_pipeline_id: 根节点 ID
        :param node_id: 节点 ID
        :param version: 节点版本
        :param action_name: Action 名称
        :return:
        """
        if action_name not in cls.__hub:
            raise ValueError("{} not found".format(action_name))
        return cls.__hub[action_name](root_pipeline_id, node_id, version)


class ActionMeta(type):
    """
    Metaclass for FEEL function invocation
    """

    def __new__(cls, name, bases, dct):
        # ensure initialization is only performed for subclasses of Plugin
        parents = [b for b in bases if isinstance(b, ActionMeta)]
        if not parents:
            return super().__new__(cls, name, bases, dct)

        new_cls = super().__new__(cls, name, bases, dct)

        # meta validation
        meta_obj = getattr(new_cls, "Meta", None)
        if not meta_obj:
            raise AttributeError("Meta class is required")

        action_name = getattr(meta_obj, "action_name", None)
        if not action_name:
            raise AttributeError("action_name is required in Meta")

        # register func
        ActionManager.register_invocation_cls(new_cls)

        return new_cls


class BaseAction(metaclass=ActionMeta):
    def __init__(self, root_pipeline_id: str, node_id: str, version: str):
        self.root_pipeline_id = root_pipeline_id
        self.node_id = node_id
        self.version = version

    @classmethod
    def get_execution_data(cls, root_pipeline_id: str, node_id: str) -> Tuple[DataObject, DataObject]:
        runtime: BambooDjangoRuntime = BambooDjangoRuntime()
        data: ExecutionData = runtime.get_execution_data(node_id)
        root_pipeline_inputs: Dict[str, Any] = {
            key: di.value for key, di in runtime.get_data_inputs(root_pipeline_id).items()
        }
        root_pipeline_data: ExecutionData = ExecutionData(inputs=root_pipeline_inputs, outputs={})

        data_obj: DataObject = DataObject(inputs=data.inputs, outputs=data.outputs)
        parent_data_obj: DataObject = DataObject(inputs=root_pipeline_data.inputs, outputs=root_pipeline_data.outputs)
        return data_obj, parent_data_obj

    def notify(self, *args, **kwargs) -> bool:
        data, parent_data = self.get_execution_data(self.root_pipeline_id, self.node_id)
        return self.do(data, parent_data, *args, **kwargs)

    @abstractmethod
    def do(self, data: DataObject, parent_data: DataObject, *args, **kwargs) -> bool:
        raise NotImplementedError


class ExampleAction(BaseAction):
    def do(self, data: DataObject, parent_data: DataObject, *args, **kwargs) -> bool:
        logger.info("[Action] example do: data -> %s, parent_data -> %s", data, parent_data)
        return True

    class Meta:
        action_name = "example"


class ForcedFailAction(BaseAction):

    TIMEOUT_NODE_OPERATOR = "bamboo_engine"

    def do(self, data: DataObject, parent_data: DataObject, *args, **kwargs) -> bool:
        logger.info("[Action(bamboo_engine_forced_fail)] do: data -> %s, parent_data -> %s", data, parent_data)
        result = bamboo_engine_api.forced_fail_activity(
            runtime=BambooDjangoRuntime(),
            node_id=self.node_id,
            ex_data="forced fail by {}".format(self.TIMEOUT_NODE_OPERATOR),
            send_post_set_state_signal=kwargs.get("send_post_set_state_signal", True),
        )
        return result.result

    class Meta:
        action_name = "bamboo_engine_forced_fail"


class ForcedFailAndSkipAction(BaseAction):

    TIMEOUT_NODE_OPERATOR = "bamboo_engine"

    def do(self, data: DataObject, parent_data: DataObject, *args, **kwargs) -> bool:
        logger.info("[Action(bamboo_engine_forced_fail_and_skip)] do: data -> %s, parent_data -> %s", data, parent_data)
        result = bamboo_engine_api.forced_fail_activity(
            runtime=BambooDjangoRuntime(),
            node_id=self.node_id,
            ex_data="forced fail by {}".format(self.TIMEOUT_NODE_OPERATOR),
            send_post_set_state_signal=kwargs.get("send_post_set_state_signal", True),
        )
        if result.result:
            result = bamboo_engine_api.skip_node(
                runtime=BambooDjangoRuntime(),
                node_id=self.node_id,
                ex_data="forced skip by {}".format(self.TIMEOUT_NODE_OPERATOR),
                send_post_set_state_signal=kwargs.get("send_post_set_state_signal", True),
            )
        return result.result

    class Meta:
        action_name = "bamboo_engine_forced_fail_and_skip"



# Generated by Django 3.2.18 on 2023-10-16 07:12

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="RollbackNodeSnapshot",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("root_pipeline_id", models.CharField(db_index=True, max_length=64, verbose_name="root pipeline id")),
                ("node_id", models.CharField(db_index=True, max_length=64, verbose_name="node_id")),
                ("code", models.CharField(max_length=64, verbose_name="node_code")),
                ("version", models.CharField(max_length=33, verbose_name="version")),
                ("inputs", models.TextField(verbose_name="node inputs")),
                ("outputs", models.TextField(verbose_name="node outputs")),
                ("context_values", models.TextField(verbose_name="pipeline context values")),
                ("rolled_back", models.BooleanField(default=False, verbose_name="whether the node rolls back")),
            ],
        ),
        migrations.CreateModel(
            name="RollbackPlan",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("root_pipeline_id", models.CharField(db_index=True, max_length=64, verbose_name="root pipeline id")),
                ("start_node_id", models.CharField(db_index=True, max_length=64, verbose_name="start node id")),
                ("target_node_id", models.CharField(db_index=True, max_length=64, verbose_name="target_node_id")),
                ("mode", models.CharField(default="TOKEN", max_length=32, verbose_name="rollback mode")),
                ("is_expired", models.BooleanField(default=False, verbose_name="is expired")),
            ],
        ),
        migrations.CreateModel(
            name="RollbackSnapshot",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("root_pipeline_id", models.CharField(db_index=True, max_length=64, verbose_name="root pipeline id")),
                ("graph", models.TextField(verbose_name="rollback graph")),
                ("node_access_record", models.TextField(verbose_name="node access record")),
                ("skip_rollback_nodes", models.TextField(verbose_name="skip rollback nodes")),
                ("other_nodes", models.TextField(verbose_name="other nodes")),
                ("start_node_id", models.CharField(db_index=True, max_length=64, verbose_name="start node id")),
                ("target_node_id", models.CharField(db_index=True, max_length=64, verbose_name="target_node_id")),
                ("is_expired", models.BooleanField(db_index=True, default=False, verbose_name="is expired")),
            ],
        ),
        migrations.CreateModel(
            name="RollbackToken",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("root_pipeline_id", models.CharField(db_index=True, max_length=64, verbose_name="root pipeline id")),
                ("token", models.TextField(verbose_name="token map")),
                (
                    "is_deleted",
                    models.BooleanField(
                        db_index=True, default=False, help_text="is deleted", verbose_name="is deleted"
                    ),
                ),
            ],
        ),
    ]
